"""Utility functions for performance monitoring."""
import os
from typing import Optional

from sky import sky_logging
from sky.skylet import constants

logger = sky_logging.init_logger(__name__)


def get_loop_lag_threshold() -> Optional[float]:
    """Get the loop lag threshold from the environment variable."""
    lag_threshold = os.getenv(constants.ENV_VAR_LOOP_LAG_THRESHOLD_MS, None)
    if lag_threshold is not None:
        try:
            return float(lag_threshold) / 1000.0
        except ValueError:
            logger.warning(
                f'Invalid value for {constants.ENV_VAR_LOOP_LAG_THRESHOLD_MS}:'
                f' {lag_threshold}')
            return None
    return None



"""State for API server process."""

# This state is used to block requests except /api operations, which is useful
# when a server is shutting down: new requests will be blocked, but existing
# requests will be allowed to finish and be operated via /api operations, e.g.
# /api/logs, /api/cancel, etc.
_block_requests = False


# TODO(aylei): refactor, state should be a instance property of API server app
# instead of a global variable.
def get_block_requests() -> bool:
    """Whether block requests except /api operations."""
    return _block_requests


def set_block_requests(shutting_down: bool) -> None:
    """Set the API server to block requests except /api operations."""
    global _block_requests
    _block_requests = shutting_down



"""GCP provisioner for SkyPilot."""

from sky.provision.runpod.config import bootstrap_instances
from sky.provision.runpod.instance import cleanup_ports
from sky.provision.runpod.instance import get_cluster_info
from sky.provision.runpod.instance import query_instances
from sky.provision.runpod.instance import query_ports
from sky.provision.runpod.instance import run_instances
from sky.provision.runpod.instance import stop_instances
from sky.provision.runpod.instance import terminate_instances
from sky.provision.runpod.instance import wait_instances
from sky.provision.runpod.volume import apply_volume
from sky.provision.runpod.volume import delete_volume
from sky.provision.runpod.volume import get_volume_usedby



"""Directory utilities."""

import os

# This file is in '<project_root>/sky/utils/directory_utils.py'
# So we need to go up 2 levels to get to the '<project_root>/sky' directory
SKY_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))


def get_sky_dir():
    """Get the sky root directory."""
    return SKY_DIR



"""Instrumentation for the API server."""

import contextlib
import functools
import multiprocessing
import os
import time

import fastapi
from prometheus_client import generate_latest
from prometheus_client import multiprocess
import prometheus_client as prom
import psutil
import starlette.middleware.base
import uvicorn

from sky import sky_logging
from sky.skylet import constants

# Whether the metrics are enabled, cannot be changed at runtime.
METRICS_ENABLED = os.environ.get(constants.ENV_VAR_SERVER_METRICS_ENABLED,
                                 'false').lower() == 'true'

logger = sky_logging.init_logger(__name__)

# Total number of API server requests, grouped by path, method, and status.
SKY_APISERVER_REQUESTS_TOTAL = prom.Counter(
    'sky_apiserver_requests_total',
    'Total number of API server requests',
    ['path', 'method', 'status'],
)

# Time spent processing API server requests, grouped by path, method, and
# status.
SKY_APISERVER_REQUEST_DURATION_SECONDS = prom.Histogram(
    'sky_apiserver_request_duration_seconds',
    'Time spent processing API server requests',
    ['path', 'method', 'status'],
    buckets=(0.01, 0.025, 0.05, 0.1, 0.25, 0.5, 1.0, 2.5, 5.0, 10.0, 20.0, 30.0,
             60.0, 120.0, float('inf')),
)

# Time spent processing a piece of code, refer to time_it().
SKY_APISERVER_CODE_DURATION_SECONDS = prom.Histogram(
    'sky_apiserver_code_duration_seconds',
    'Time spent processing code',
    ['name', 'group'],
    buckets=(0.01, 0.025, 0.05, 0.1, 0.25, 0.5, 1.0, 2.5, 5.0, 10.0, 20.0, 30.0,
             60.0, 120.0, float('inf')),
)

SKY_APISERVER_EVENT_LOOP_LAG_SECONDS = prom.Histogram(
    'sky_apiserver_event_loop_lag_seconds',
    'Scheduling delay of the server event loop',
    ['pid'],
    buckets=(0.001, 0.005, 0.01, 0.025, 0.05, 0.1, 0.25, 0.5, 1, 2, 5, 20.0,
             60.0, float('inf')),
)

SKY_APISERVER_WEBSOCKET_CONNECTIONS = prom.Gauge(
    'sky_apiserver_websocket_connections',
    'Number of websocket connections',
    ['pid'],
    multiprocess_mode='livesum',
)

SKY_APISERVER_WEBSOCKET_CLOSED_TOTAL = prom.Counter(
    'sky_apiserver_websocket_closed_total',
    'Number of websocket closed',
    ['pid', 'reason'],
)

# The number of execution starts in each worker process, we do not record
# histogram here as the duration has been measured in
# SKY_APISERVER_CODE_DURATION_SECONDS without the worker label (process id).
# Recording histogram WITH worker label will cause high cardinality.
SKY_APISERVER_PROCESS_EXECUTION_START_TOTAL = prom.Counter(
    'sky_apiserver_process_execution_start_total',
    'Total number of execution starts in each worker process',
    ['request', 'pid'],
)

SKY_APISERVER_PROCESS_PEAK_RSS = prom.Gauge(
    'sky_apiserver_process_peak_rss',
    'Peak RSS we saw in each process in last 30 seconds',
    ['pid', 'type'],
)

SKY_APISERVER_PROCESS_CPU_TOTAL = prom.Gauge(
    'sky_apiserver_process_cpu_total',
    'Total CPU times a worker process has been running',
    ['pid', 'type', 'mode'],
)

metrics_app = fastapi.FastAPI()


@metrics_app.get('/metrics')
async def metrics() -> fastapi.Response:
    """Expose aggregated Prometheus metrics from all worker processes."""
    if os.environ.get('PROMETHEUS_MULTIPROC_DIR'):
        # In multiprocess mode, we need to collect metrics from all processes.
        registry = prom.CollectorRegistry()
        multiprocess.MultiProcessCollector(registry)
        data = generate_latest(registry)
    else:
        data = generate_latest()
    return fastapi.Response(content=data,
                            media_type=prom.CONTENT_TYPE_LATEST,
                            headers={'Cache-Control': 'no-cache'})


def build_metrics_server(host: str, port: int) -> uvicorn.Server:
    metrics_config = uvicorn.Config(
        'sky.server.metrics:metrics_app',
        host=host,
        port=port,
        workers=1,
    )
    metrics_server_instance = uvicorn.Server(metrics_config)
    return metrics_server_instance


def _get_status_code_group(status_code: int) -> str:
    """Group status codes into classes (2xx, 5xx) to reduce cardinality."""
    return f'{status_code // 100}xx'


def _is_streaming_api(path: str) -> bool:
    """Check if the path is a streaming API."""
    path = path.rstrip('/')
    return path.endswith('/logs') or path.endswith('/api/stream')


class PrometheusMiddleware(starlette.middleware.base.BaseHTTPMiddleware):
    """Middleware to collect Prometheus metrics for HTTP requests."""

    async def dispatch(self, request: fastapi.Request, call_next):
        path = request.url.path
        logger.debug(f'PROM Middleware Request: {request}, {request.url.path}')
        streaming = _is_streaming_api(path)
        if not streaming:
            # Exclude streaming APIs, the duration is not meaningful.
            # TODO(aylei): measure the duration of async execution instead.
            start_time = time.time()
        method = request.method
        status_code_group = ''

        try:
            response = await call_next(request)
            status_code_group = _get_status_code_group(response.status_code)
        except Exception:  # pylint: disable=broad-except
            status_code_group = '5xx'
            raise
        finally:
            SKY_APISERVER_REQUESTS_TOTAL.labels(path=path,
                                                method=method,
                                                status=status_code_group).inc()
            if not streaming:
                duration = time.time() - start_time
                SKY_APISERVER_REQUEST_DURATION_SECONDS.labels(
                    path=path, method=method,
                    status=status_code_group).observe(duration)

        return response


@contextlib.contextmanager
def time_it(name: str, group: str = 'default'):
    """Context manager to measure and record code execution duration."""
    if not METRICS_ENABLED:
        yield
    else:
        start_time = time.time()
        try:
            yield
        finally:
            duration = time.time() - start_time
            SKY_APISERVER_CODE_DURATION_SECONDS.labels(
                name=name, group=group).observe(duration)


def time_me(func):
    """Measure the duration of decorated function."""

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        if not METRICS_ENABLED:
            return func(*args, **kwargs)
        name = f'{func.__module__}/{func.__name__}'
        with time_it(name, group='function'):
            return func(*args, **kwargs)

    return wrapper


def time_me_async(func):
    """Measure the duration of decorated async function."""

    @functools.wraps(func)
    async def async_wrapper(*args, **kwargs):
        if not METRICS_ENABLED:
            return await func(*args, **kwargs)
        name = f'{func.__module__}/{func.__name__}'
        with time_it(name, group='function'):
            return await func(*args, **kwargs)

    return async_wrapper


def process_monitor(process_type: str):
    pid = multiprocessing.current_process().pid
    proc = psutil.Process(pid)
    peak_rss = 0
    last_bucket_end = time.time()
    while True:
        if time.time() - last_bucket_end >= 30:
            # Reset peak RSS every 30 seconds.
            last_bucket_end = time.time()
            peak_rss = 0
        peak_rss = max(peak_rss, proc.memory_info().rss)
        SKY_APISERVER_PROCESS_PEAK_RSS.labels(pid=pid,
                                              type=process_type).set(peak_rss)
        ctimes = proc.cpu_times()
        SKY_APISERVER_PROCESS_CPU_TOTAL.labels(pid=pid,
                                               type=process_type,
                                               mode='user').set(ctimes.user)
        SKY_APISERVER_PROCESS_CPU_TOTAL.labels(pid=pid,
                                               type=process_type,
                                               mode='system').set(ctimes.system)
        time.sleep(1)



"""DO provisioner for SkyPilot."""

from sky.provision.do.config import bootstrap_instances
from sky.provision.do.instance import cleanup_ports
from sky.provision.do.instance import get_cluster_info
from sky.provision.do.instance import open_ports
from sky.provision.do.instance import query_instances
from sky.provision.do.instance import run_instances
from sky.provision.do.instance import stop_instances
from sky.provision.do.instance import terminate_instances
from sky.provision.do.instance import wait_instances



"""SSH Catalog.

This catalog inherits from the Kubernetes catalog as SSH cloud is a wrapper
around Kubernetes that uses SSH-specific contexts.
"""
import typing
from typing import Dict, List, Optional, Tuple

from sky import sky_logging
from sky.catalog import CloudFilter
from sky.catalog import common
from sky.catalog import kubernetes_catalog
from sky.clouds import ssh

logger = sky_logging.init_logger(__name__)

if typing.TYPE_CHECKING:
    import pandas as pd
else:
    from sky.adaptors import common as adaptors_common
    pd = adaptors_common.LazyImport('pandas')

_PULL_FREQUENCY_HOURS = 7

# Reuse the Kubernetes images catalog for SSH cloud.
# We keep pull_frequency_hours so we can remotely update the default image paths
_image_df = common.read_catalog('kubernetes/images.csv',
                                pull_frequency_hours=_PULL_FREQUENCY_HOURS)


def get_image_id_from_tag(tag: str, region: Optional[str]) -> Optional[str]:
    """Returns the image id from the tag.

    Delegates to Kubernetes catalog implementation.
    """
    return kubernetes_catalog.get_image_id_from_tag(tag, region)


def is_image_tag_valid(tag: str, region: Optional[str]) -> bool:
    """Returns whether the image tag is valid.

    Delegates to Kubernetes catalog implementation.
    """
    return kubernetes_catalog.is_image_tag_valid(tag, region)


def list_accelerators(
        gpus_only: bool,
        name_filter: Optional[str],
        region_filter: Optional[str],
        quantity_filter: Optional[int],
        case_sensitive: bool = True,
        all_regions: bool = False,
        require_price: bool = True) -> Dict[str, List[common.InstanceTypeInfo]]:
    """List accelerators in SSH-based Kubernetes clusters.

    Delegates to the Kubernetes _list_accelerators function but restricts to
    SSH contexts.
    """
    return _list_accelerators(gpus_only,
                              name_filter,
                              region_filter,
                              quantity_filter,
                              case_sensitive,
                              all_regions,
                              require_price,
                              realtime=False)[0]


def list_accelerators_realtime(
    gpus_only: bool,
    name_filter: Optional[str],
    region_filter: Optional[str],
    quantity_filter: Optional[int],
    case_sensitive: bool = True,
    all_regions: bool = False,
    require_price: bool = True
) -> Tuple[Dict[str, List[common.InstanceTypeInfo]], Dict[str, int], Dict[str,
                                                                          int]]:
    """List accelerators in SSH Node Pools with real-time information.

    Delegates to the Kubernetes _list_accelerators function but restricts to
    SSH contexts.
    """
    return _list_accelerators(gpus_only,
                              name_filter,
                              region_filter,
                              quantity_filter,
                              case_sensitive,
                              all_regions,
                              require_price,
                              realtime=True)


def _list_accelerators(
    gpus_only: bool,
    name_filter: Optional[str],
    region_filter: Optional[str],
    quantity_filter: Optional[int],
    case_sensitive: bool = True,
    all_regions: bool = False,
    require_price: bool = True,
    realtime: bool = False
) -> Tuple[Dict[str, List[common.InstanceTypeInfo]], Dict[str, int], Dict[str,
                                                                          int]]:
    """List accelerators in SSH-based Kubernetes clusters.

    This is a wrapper around the Kubernetes _list_accelerators function that
    restricts the contexts to SSH-specific contexts only.

    If region_filter is specified and it's not an SSH context, no results will
    be returned.
    """
    # If a specific region is requested, ensure it's an SSH context
    if region_filter is not None and not region_filter.startswith('ssh-'):
        return {}, {}, {}

    # Get SSH contexts
    ssh_contexts = ssh.SSH.existing_allowed_contexts()

    # If no contexts found, return empty results
    if not ssh_contexts:
        return {}, {}, {}

    # If a region filter is specified and it's not a SSH context return empty
    # results
    if region_filter is not None and region_filter not in ssh_contexts:
        return {}, {}, {}

    # If region_filter is None, use the first context if all_regions is False
    if region_filter is None and not all_regions and ssh_contexts:
        # Use the first SSH context if no specific region requested
        region_filter = ssh_contexts[0]

    # Call the Kubernetes _list_accelerators with the appropriate region filter
    if realtime:
        return kubernetes_catalog.list_accelerators_realtime(
            gpus_only, name_filter, region_filter, quantity_filter,
            case_sensitive, all_regions, require_price)
    else:
        result = kubernetes_catalog.list_accelerators(
            gpus_only, name_filter, region_filter, quantity_filter,
            case_sensitive, all_regions, require_price)
        return result, {}, {}


def validate_region_zone(
        region_name: Optional[str],
        zone_name: Optional[str],
        clouds: CloudFilter = None) -> Tuple[Optional[str], Optional[str]]:
    """Validates the region and zone for SSH cloud.

    Delegates to the Kubernetes catalog implementation but ensures
    the region is a valid SSH context.
    """
    # Delegate to Kubernetes implementation
    region, zone = kubernetes_catalog.validate_region_zone(
        region_name, zone_name, clouds)

    # Get SSH contexts
    ssh_contexts = ssh.SSH.existing_allowed_contexts()

    # If a region is specified, ensure it's in the list of SSH contexts
    if region is not None and region not in ssh_contexts:
        return None, None

    return region, zone



"""Hyperbolic provisioner for SkyPilot."""

from sky.provision.hyperbolic.config import bootstrap_instances
from sky.provision.hyperbolic.instance import cleanup_custom_multi_network
from sky.provision.hyperbolic.instance import cleanup_ports
from sky.provision.hyperbolic.instance import get_cluster_info
from sky.provision.hyperbolic.instance import open_ports
from sky.provision.hyperbolic.instance import query_instances
from sky.provision.hyperbolic.instance import run_instances
from sky.provision.hyperbolic.instance import stop_instances
from sky.provision.hyperbolic.instance import terminate_instances
from sky.provision.hyperbolic.instance import wait_instances



"""gRPC service implementations for skylet."""

import grpc

from sky import sky_logging
from sky.schemas.generated import autostopv1_pb2
from sky.schemas.generated import autostopv1_pb2_grpc
from sky.skylet import autostop_lib

logger = sky_logging.init_logger(__name__)


class AutostopServiceImpl(autostopv1_pb2_grpc.AutostopServiceServicer):
    """Implementation of the AutostopService gRPC service."""

    def SetAutostop(  # type: ignore[return]
            self, request: autostopv1_pb2.SetAutostopRequest,
            context: grpc.ServicerContext
    ) -> autostopv1_pb2.SetAutostopResponse:
        """Sets autostop configuration for the cluster."""
        try:
            wait_for = autostop_lib.AutostopWaitFor.from_protobuf(
                request.wait_for)
            autostop_lib.set_autostop(
                idle_minutes=request.idle_minutes,
                backend=request.backend,
                wait_for=wait_for if wait_for is not None else
                autostop_lib.DEFAULT_AUTOSTOP_WAIT_FOR,
                down=request.down)
            return autostopv1_pb2.SetAutostopResponse()
        except Exception as e:  # pylint: disable=broad-except
            context.abort(grpc.StatusCode.INTERNAL, str(e))

    def IsAutostopping(  # type: ignore[return]
            self, request: autostopv1_pb2.IsAutostoppingRequest,
            context: grpc.ServicerContext
    ) -> autostopv1_pb2.IsAutostoppingResponse:
        """Checks if the cluster is currently autostopping."""
        try:
            is_autostopping = autostop_lib.get_is_autostopping()
            return autostopv1_pb2.IsAutostoppingResponse(
                is_autostopping=is_autostopping)
        except Exception as e:  # pylint: disable=broad-except
            context.abort(grpc.StatusCode.INTERNAL, str(e))



"""Sky logging agents."""
from typing import Optional

from sky import exceptions
from sky import skypilot_config
from sky.logs.agent import LoggingAgent
from sky.logs.aws import CloudwatchLoggingAgent
from sky.logs.gcp import GCPLoggingAgent


def get_logging_agent() -> Optional[LoggingAgent]:
    store = skypilot_config.get_nested(('logs', 'store'), None)
    if store is None:
        return None
    if store == 'gcp':
        return GCPLoggingAgent(skypilot_config.get_nested(('logs', 'gcp'), {}))
    elif store == 'aws':
        return CloudwatchLoggingAgent(
            skypilot_config.get_nested(('logs', 'aws'), {}))
    raise exceptions.InvalidSkyPilotConfigError(
        f'Invalid logging store: {store}')



"""Helper script to strip path information from exec auth in a kubeconfig file.

This script processes a kubeconfig file and removes any path information from
the 'command' field in the exec configuration, leaving only the executable name.
This is useful when moving between different environments where auth plugin
executables might be installed in different locations.

For Nebius kubeconfigs, it also changes the --profile argument to 'sky' to
ensure compatibility with SkyPilot's expected profile configuration.

It assumes the target environment has the auth executable available in PATH.
If not, you'll need to update your environment container to include the auth
executable in PATH.

When using LOCAL_CREDENTIALS (aka exec auth) with Kubernetes, though, SkyPilot
will automatically inject a wrapper script for common exec auth providers like
GKE and EKS. This wrapper script helps to resolve path issues that may arise
from executables installed on non system-default paths. Thus, the kubeconfig
file may look different on the sky jobs controller.

Usage:
    python -m sky.utils.kubernetes.exec_kubeconfig_converter
"""
import argparse
import os

import yaml

from sky.provision.kubernetes import utils as kubernetes_utils


def main():
    parser = argparse.ArgumentParser(
        description='Strip path information from exec plugin commands in a '
        'kubeconfig file. Used to prepare kubeconfigs for deployment '
        'with SkyPilot.')
    parser.add_argument(
        '--input',
        '-i',
        default=os.path.expanduser('~/.kube/config'),
        help='Input kubeconfig file path (default: %(default)s)')
    parser.add_argument(
        '--output',
        '-o',
        default=os.path.expanduser('~/.kube/config.converted'),
        help='Output kubeconfig file path (default: %(default)s)')

    args = parser.parse_args()

    with open(args.input, 'r', encoding='utf-8') as file:
        config = yaml.safe_load(file)

    updated = kubernetes_utils.format_kubeconfig_exec_auth(
        config, args.output, False)

    if updated:
        print('Kubeconfig updated with path-less exec auth. '
              f'Saved to {args.output}')
    else:
        print('No updates made.')


if __name__ == '__main__':
    main()



""" Vast Cloud. """

import os
import typing
from typing import Dict, Iterator, List, Optional, Tuple, Union

from sky import catalog
from sky import clouds
from sky.adaptors import common
from sky.utils import registry
from sky.utils import resources_utils

if typing.TYPE_CHECKING:
    from sky import resources as resources_lib
    from sky.utils import volume as volume_lib

_CREDENTIAL_PATH = '~/.config/vastai/vast_api_key'


@registry.CLOUD_REGISTRY.register
class Vast(clouds.Cloud):
    """ Vast GPU Cloud

    _REPR | The string representation for the Vast GPU cloud object.
    """
    _REPR = 'Vast'
    _CLOUD_UNSUPPORTED_FEATURES = {
        clouds.CloudImplementationFeatures.MULTI_NODE:
            ('Multi-node not supported yet, as the interconnection among nodes '
             'are non-trivial on Vast.'),
        clouds.CloudImplementationFeatures.CUSTOM_DISK_TIER:
            ('Customizing disk tier is not supported yet on Vast.'),
        clouds.CloudImplementationFeatures.CUSTOM_NETWORK_TIER:
            ('Custom network tier is currently not supported in '
             f'{_REPR}.'),
        clouds.CloudImplementationFeatures.STORAGE_MOUNTING:
            ('Mounting object stores is not supported on Vast.'),
        clouds.CloudImplementationFeatures.HIGH_AVAILABILITY_CONTROLLERS:
            ('High availability controllers are not supported on Vast.'),
        clouds.CloudImplementationFeatures.CUSTOM_MULTI_NETWORK:
            ('Customized multiple network interfaces are not supported on Vast.'
            ),
    }
    #
    # Vast doesn't have a max cluster name limit. This number
    # is reasonably large and exists to play nicely with the
    # other providers
    #
    _MAX_CLUSTER_NAME_LEN_LIMIT = 120
    _regions: List[clouds.Region] = []

    PROVISIONER_VERSION = clouds.ProvisionerVersion.SKYPILOT
    STATUS_VERSION = clouds.StatusVersion.SKYPILOT
    OPEN_PORTS_VERSION = clouds.OpenPortsVersion.LAUNCH_ONLY

    @classmethod
    def _unsupported_features_for_resources(
        cls, resources: 'resources_lib.Resources'
    ) -> Dict[clouds.CloudImplementationFeatures, str]:
        """The features not supported based on the resources provided.

        This method is used by check_features_are_supported() to check if the
        cloud implementation supports all the requested features.

        Returns:
            A dict of {feature: reason} for the features not supported by the
            cloud implementation.
        """
        del resources  # unused
        return cls._CLOUD_UNSUPPORTED_FEATURES

    @classmethod
    def _max_cluster_name_length(cls) -> Optional[int]:
        return cls._MAX_CLUSTER_NAME_LEN_LIMIT

    @classmethod
    def regions_with_offering(cls, instance_type: str,
                              accelerators: Optional[Dict[str, int]],
                              use_spot: bool, region: Optional[str],
                              zone: Optional[str]) -> List[clouds.Region]:
        assert zone is None, 'Vast does not support zones.'
        del accelerators, zone  # unused
        regions = catalog.get_region_zones_for_instance_type(
            instance_type, use_spot, 'vast')

        if region is not None:
            regions = [r for r in regions if r.name == region]
        return regions

    @classmethod
    def get_vcpus_mem_from_instance_type(
        cls,
        instance_type: str,
    ) -> Tuple[Optional[float], Optional[float]]:
        return catalog.get_vcpus_mem_from_instance_type(instance_type,
                                                        clouds='vast')

    @classmethod
    def zones_provision_loop(
        cls,
        *,
        region: str,
        num_nodes: int,
        instance_type: str,
        accelerators: Optional[Dict[str, int]] = None,
        use_spot: bool = False,
    ) -> Iterator[None]:
        del num_nodes  # unused
        regions = cls.regions_with_offering(instance_type,
                                            accelerators,
                                            use_spot,
                                            region=region,
                                            zone=None)
        for r in regions:
            assert r.zones is None, r
            yield r.zones

    def instance_type_to_hourly_cost(self,
                                     instance_type: str,
                                     use_spot: bool,
                                     region: Optional[str] = None,
                                     zone: Optional[str] = None) -> float:
        return catalog.get_hourly_cost(instance_type,
                                       use_spot=use_spot,
                                       region=region,
                                       zone=zone,
                                       clouds='vast')

    def accelerators_to_hourly_cost(self,
                                    accelerators: Dict[str, int],
                                    use_spot: bool,
                                    region: Optional[str] = None,
                                    zone: Optional[str] = None) -> float:
        """Returns the hourly cost of the accelerators, in dollars/hour."""
        del accelerators, use_spot, region, zone  # unused
        return 0.0  # Vast includes accelerators in the hourly cost.

    def get_egress_cost(self, num_gigabytes: float) -> float:
        return 0.0

    @classmethod
    def get_default_instance_type(cls,
                                  cpus: Optional[str] = None,
                                  memory: Optional[str] = None,
                                  disk_tier: Optional[
                                      resources_utils.DiskTier] = None,
                                  region: Optional[str] = None,
                                  zone: Optional[str] = None) -> Optional[str]:
        """Returns the default instance type for Vast."""
        return catalog.get_default_instance_type(cpus=cpus,
                                                 memory=memory,
                                                 disk_tier=disk_tier,
                                                 region=region,
                                                 zone=zone,
                                                 clouds='vast')

    @classmethod
    def get_accelerators_from_instance_type(
            cls, instance_type: str) -> Optional[Dict[str, Union[int, float]]]:
        return catalog.get_accelerators_from_instance_type(instance_type,
                                                           clouds='vast')

    @classmethod
    def get_zone_shell_cmd(cls) -> Optional[str]:
        return None

    def make_deploy_resources_variables(
        self,
        resources: 'resources_lib.Resources',
        cluster_name: resources_utils.ClusterName,
        region: 'clouds.Region',
        zones: Optional[List['clouds.Zone']],
        num_nodes: int,
        dryrun: bool = False,
        volume_mounts: Optional[List['volume_lib.VolumeMount']] = None,
    ) -> Dict[str, Optional[str]]:
        del zones, dryrun, cluster_name, num_nodes  # unused

        resources = resources.assert_launchable()
        acc_dict = self.get_accelerators_from_instance_type(
            resources.instance_type)
        custom_resources = resources_utils.make_ray_custom_resources_str(
            acc_dict)

        if resources.image_id is None:
            image_id: Optional[str] = 'vastai/base:0.0.2'
        elif resources.extract_docker_image() is not None:
            image_id = resources.extract_docker_image()
        else:
            image_id = resources.image_id[resources.region]

        return {
            'instance_type': resources.instance_type,
            'custom_resources': custom_resources,
            'region': region.name,
            'image_id': image_id,
        }

    def _get_feasible_launchable_resources(
        self, resources: 'resources_lib.Resources'
    ) -> 'resources_utils.FeasibleResources':
        """Returns a list of feasible resources for the given resources."""
        if resources.instance_type is not None:
            assert resources.is_launchable(), resources
            resources = resources.copy(accelerators=None)
            return resources_utils.FeasibleResources([resources], [], None)

        def _make(instance_list):
            resource_list = []
            for instance_type in instance_list:
                r = resources.copy(
                    cloud=Vast(),
                    instance_type=instance_type,
                    accelerators=None,
                    cpus=None,
                )
                resource_list.append(r)
            return resource_list

        # Currently, handle a filter on accelerators only.
        accelerators = resources.accelerators
        if accelerators is None:
            # Return a default instance type
            default_instance_type = Vast.get_default_instance_type(
                cpus=resources.cpus,
                memory=resources.memory,
                disk_tier=resources.disk_tier,
                region=resources.region,
                zone=resources.zone)
            if default_instance_type is None:
                # TODO: Add hints to all return values in this method to help
                #  users understand why the resources are not launchable.
                return resources_utils.FeasibleResources([], [], None)
            else:
                return resources_utils.FeasibleResources(
                    _make([default_instance_type]), [], None)

        assert len(accelerators) == 1, resources
        acc, acc_count = list(accelerators.items())[0]
        (instance_list,
         fuzzy_candidate_list) = catalog.get_instance_type_for_accelerator(
             acc,
             acc_count,
             use_spot=resources.use_spot,
             cpus=resources.cpus,
             region=resources.region,
             zone=resources.zone,
             memory=resources.memory,
             clouds='vast')
        if instance_list is None:
            return resources_utils.FeasibleResources([], fuzzy_candidate_list,
                                                     None)
        return resources_utils.FeasibleResources(_make(instance_list),
                                                 fuzzy_candidate_list, None)

    @classmethod
    def _check_compute_credentials(
            cls) -> Tuple[bool, Optional[Union[str, Dict[str, str]]]]:
        """Checks if the user has valid credentials for
        Vast's compute service."""

        dependency_error_msg = ('Failed to import vast. '
                                'To install, run: pip install skypilot[vast]')
        if not common.can_import_modules(['vastai_sdk']):
            return False, dependency_error_msg

        if not os.path.exists(os.path.expanduser(_CREDENTIAL_PATH)):
            return False, (
                'error \n'  # First line is indented by 4 spaces
                '    Credentials can be set up by running: \n'
                '        $ pip install vastai\n'
                '        $ mkdir -p ~/.config/vastai\n'
                f'        $ echo [key] > {_CREDENTIAL_PATH}\n'
                '    For more information, see https://skypilot.readthedocs.io/en/latest/getting-started/installation.html#vast'  # pylint: disable=line-too-long
            )

        return True, None

    def get_credential_file_mounts(self) -> Dict[str, str]:
        return {f'{_CREDENTIAL_PATH}': f'{_CREDENTIAL_PATH}'}

    @classmethod
    def get_user_identities(cls) -> Optional[List[List[str]]]:
        # NOTE: used for very advanced SkyPilot functionality
        # Can implement later if desired
        return None

    def instance_type_exists(self, instance_type: str) -> bool:
        return catalog.instance_type_exists(instance_type, 'vast')

    def validate_region_zone(self, region: Optional[str], zone: Optional[str]):
        return catalog.validate_region_zone(region, zone, clouds='vast')

    @classmethod
    def get_image_size(cls, image_id: str, region: Optional[str]) -> float:
        # TODO: use 0.0 for now to allow all images. We should change this to
        # return the docker image size.
        return 0.0



"""Volume management core."""

import contextlib
import os
from typing import Any, Dict, Generator, List, Optional
import uuid

import filelock

from sky import global_user_state
from sky import models
from sky import provision
from sky import sky_logging
from sky.utils import common_utils
from sky.utils import registry
from sky.utils import rich_utils
from sky.utils import status_lib
from sky.utils import ux_utils

logger = sky_logging.init_logger(__name__)

# Filelocks for the storage management.
VOLUME_LOCK_PATH = os.path.expanduser('~/.sky/.{volume_name}.lock')
VOLUME_LOCK_TIMEOUT_SECONDS = 20


def volume_refresh():
    """Refreshes the volume status."""
    volumes = global_user_state.get_volumes()
    for volume in volumes:
        volume_name = volume.get('name')
        config = volume.get('handle')
        if config is None:
            logger.warning(f'Volume {volume_name} has no handle.'
                           'Skipping status refresh...')
            continue
        cloud = config.cloud
        usedby_pods, _ = provision.get_volume_usedby(cloud, config)
        with _volume_lock(volume_name):
            latest_volume = global_user_state.get_volume_by_name(volume_name)
            if latest_volume is None:
                logger.warning(f'Volume {volume_name} not found.')
                continue
            status = latest_volume.get('status')
            if not usedby_pods:
                if status != status_lib.VolumeStatus.READY:
                    logger.info(f'Update volume {volume_name} '
                                f'status to READY')
                    global_user_state.update_volume_status(
                        volume_name, status=status_lib.VolumeStatus.READY)
            else:
                if status != status_lib.VolumeStatus.IN_USE:
                    logger.info(f'Update volume {volume_name} '
                                f'status to IN_USE, usedby: {usedby_pods}')
                    global_user_state.update_volume_status(
                        volume_name, status=status_lib.VolumeStatus.IN_USE)


def volume_list() -> List[Dict[str, Any]]:
    """Gets the volumes.

    Returns:
        [
            {
                'name': str,
                'type': str,
                'launched_at': int timestamp of creation,
                'cloud': str,
                'region': str,
                'zone': str,
                'size': str,
                'config': Dict[str, Any],
                'name_on_cloud': str,
                'user_hash': str,
                'workspace': str,
                'last_attached_at': int timestamp of last attachment,
                'last_use': last command,
                'status': sky.VolumeStatus,
                'usedby_pods': List[str],
                'usedby_clusters': List[str],
            }
        ]
    """
    with rich_utils.safe_status(ux_utils.spinner_message('Listing volumes')):
        volumes = global_user_state.get_volumes()
        cloud_to_configs: Dict[str, List[models.VolumeConfig]] = {}
        for volume in volumes:
            config = volume.get('handle')
            if config is None:
                volume_name = volume.get('name')
                logger.warning(f'Volume {volume_name} has no handle.')
                continue
            cloud = config.cloud
            if cloud not in cloud_to_configs:
                cloud_to_configs[cloud] = []
            cloud_to_configs[cloud].append(config)

        cloud_to_used_by_pods, cloud_to_used_by_clusters = {}, {}
        for cloud, configs in cloud_to_configs.items():
            used_by_pods, used_by_clusters = provision.get_all_volumes_usedby(
                cloud, configs)
            cloud_to_used_by_pods[cloud] = used_by_pods
            cloud_to_used_by_clusters[cloud] = used_by_clusters

        all_users = global_user_state.get_all_users()
        user_map = {user.id: user.name for user in all_users}
        records = []
        for volume in volumes:
            volume_name = volume.get('name')
            record = {
                'name': volume_name,
                'launched_at': volume.get('launched_at'),
                'user_hash': volume.get('user_hash'),
                'user_name': user_map.get(volume.get('user_hash'), ''),
                'workspace': volume.get('workspace'),
                'last_attached_at': volume.get('last_attached_at'),
                'last_use': volume.get('last_use'),
                'usedby_pods': [],
                'usedby_clusters': [],
            }
            status = volume.get('status')
            if status is not None:
                record['status'] = status.value
            else:
                record['status'] = ''
            config = volume.get('handle')
            if config is None:
                logger.warning(f'Volume {volume_name} has no handle.')
                continue
            cloud = config.cloud
            usedby_pods, usedby_clusters = provision.map_all_volumes_usedby(
                cloud,
                cloud_to_used_by_pods[cloud],
                cloud_to_used_by_clusters[cloud],
                config,
            )
            record['type'] = config.type
            record['cloud'] = config.cloud
            record['region'] = config.region
            record['zone'] = config.zone
            record['size'] = config.size
            record['config'] = config.config
            record['name_on_cloud'] = config.name_on_cloud
            record['usedby_pods'] = usedby_pods
            record['usedby_clusters'] = usedby_clusters
            records.append(record)
        return records


def volume_delete(names: List[str]) -> None:
    """Deletes volumes.

    Args:
        names: List of volume names to delete.

    Raises:
        ValueError: If the volume does not exist
          or is in use or has no handle.
    """
    with rich_utils.safe_status(ux_utils.spinner_message('Deleting volumes')):
        for name in names:
            volume = global_user_state.get_volume_by_name(name)
            if volume is None:
                raise ValueError(f'Volume {name} not found.')
            config = volume.get('handle')
            if config is None:
                raise ValueError(f'Volume {name} has no handle.')
            cloud = config.cloud
            usedby_pods, usedby_clusters = provision.get_volume_usedby(
                cloud, config)
            if usedby_clusters:
                usedby_clusters_str = ', '.join(usedby_clusters)
                cluster_str = 'clusters' if len(
                    usedby_clusters) > 1 else 'cluster'
                raise ValueError(f'Volume {name} is used by {cluster_str}'
                                 f' {usedby_clusters_str}.')
            if usedby_pods:
                usedby_pods_str = ', '.join(usedby_pods)
                pod_str = 'pods' if len(usedby_pods) > 1 else 'pod'
                raise ValueError(
                    f'Volume {name} is used by {pod_str} {usedby_pods_str}.')
            logger.debug(f'Deleting volume {name} with config {config}')
            with _volume_lock(name):
                provision.delete_volume(cloud, config)
                global_user_state.delete_volume(name)


def volume_apply(
    name: str,
    volume_type: str,
    cloud: str,
    region: Optional[str],
    zone: Optional[str],
    size: Optional[str],
    config: Dict[str, Any],
    labels: Optional[Dict[str, str]] = None,
) -> None:
    """Creates or registers a volume.

    Args:
        name: The name of the volume.
        volume_type: The type of the volume.
        cloud: The cloud of the volume.
        region: The region of the volume.
        zone: The zone of the volume.
        size: The size of the volume.
        config: The configuration of the volume.
        labels: The labels of the volume.

    """
    with rich_utils.safe_status(ux_utils.spinner_message('Creating volume')):
        # Reuse the method for cluster name on cloud to
        # generate the storage name on cloud.
        cloud_obj = registry.CLOUD_REGISTRY.from_str(cloud)
        assert cloud_obj is not None
        name_uuid = str(uuid.uuid4())[:6]
        name_on_cloud = common_utils.make_cluster_name_on_cloud(
            name, max_length=cloud_obj.max_cluster_name_length())
        name_on_cloud += '-' + name_uuid
        config = models.VolumeConfig(
            name=name,
            type=volume_type,
            cloud=str(cloud_obj),
            region=region,
            zone=zone,
            size=size,
            config=config,
            name_on_cloud=name_on_cloud,
            labels=labels,
        )
        logger.debug(
            f'Creating volume {name} on cloud {cloud} with config {config}')
        with _volume_lock(name):
            current_volume = global_user_state.get_volume_by_name(name)
            if current_volume is not None:
                logger.info(f'Volume {name} already exists.')
                return
            config = provision.apply_volume(cloud, config)
            global_user_state.add_volume(name, config,
                                         status_lib.VolumeStatus.READY)


@contextlib.contextmanager
def _volume_lock(volume_name: str) -> Generator[None, None, None]:
    """Context manager for volume lock."""
    try:
        with filelock.FileLock(VOLUME_LOCK_PATH.format(volume_name=volume_name),
                               VOLUME_LOCK_TIMEOUT_SECONDS):
            yield
    except filelock.Timeout as e:
        raise RuntimeError(
            f'Failed to update user due to a timeout '
            f'when trying to acquire the lock at '
            f'{VOLUME_LOCK_PATH.format(volume_name=volume_name)}. '
            'Please try again or manually remove the lock '
            f'file if you believe it is stale.') from e



"""Service specification for SkyServe."""
import json
import os
import textwrap
from typing import Any, Dict, List, Optional, Union

from sky import serve
from sky.serve import constants
from sky.serve import load_balancing_policies as lb_policies
from sky.serve import serve_utils
from sky.serve import spot_placer as spot_placer_lib
from sky.utils import common_utils
from sky.utils import schemas
from sky.utils import ux_utils
from sky.utils import yaml_utils


class SkyServiceSpec:
    """SkyServe service specification."""

    def __init__(
        self,
        readiness_path: str,
        initial_delay_seconds: int,
        readiness_timeout_seconds: int,
        min_replicas: int,
        max_replicas: Optional[int] = None,
        num_overprovision: Optional[int] = None,
        ports: Optional[str] = None,
        target_qps_per_replica: Optional[Union[float, Dict[str, float]]] = None,
        post_data: Optional[Dict[str, Any]] = None,
        tls_credential: Optional[serve_utils.TLSCredential] = None,
        readiness_headers: Optional[Dict[str, str]] = None,
        dynamic_ondemand_fallback: Optional[bool] = None,
        base_ondemand_fallback_replicas: Optional[int] = None,
        spot_placer: Optional[str] = None,
        upscale_delay_seconds: Optional[int] = None,
        downscale_delay_seconds: Optional[int] = None,
        load_balancing_policy: Optional[str] = None,
        pool: Optional[bool] = None,
    ) -> None:
        if pool:
            for unsupported_field in [
                    'max_replicas',
                    'num_overprovision',
                    'target_qps_per_replica',
                    'upscale_delay_seconds',
                    'downscale_delay_seconds',
                    'base_ondemand_fallback_replicas',
                    'dynamic_ondemand_fallback',
                    'spot_placer',
                    'load_balancing_policy',
                    'ports',
                    'post_data',
                    'tls_credential',
                    'readiness_headers',
            ]:
                if locals()[unsupported_field] is not None:
                    with ux_utils.print_exception_no_traceback():
                        raise ValueError(
                            f'{unsupported_field} is not supported for pool.')
            if max_replicas is not None and max_replicas != min_replicas:
                with ux_utils.print_exception_no_traceback():
                    raise ValueError('Autoscaling is not supported for pool '
                                     'for now.')

        if max_replicas is not None and max_replicas < min_replicas:
            with ux_utils.print_exception_no_traceback():
                raise ValueError('max_replicas must be greater than or '
                                 'equal to min_replicas. Found: '
                                 f'min_replicas={min_replicas}, '
                                 f'max_replicas={max_replicas}')

        if target_qps_per_replica is not None:
            if max_replicas is None:
                with ux_utils.print_exception_no_traceback():
                    raise ValueError('max_replicas must be set where '
                                     'target_qps_per_replica is set.')
        else:
            if max_replicas is not None and max_replicas != min_replicas:
                with ux_utils.print_exception_no_traceback():
                    raise ValueError(
                        'Detected different min_replicas and max_replicas '
                        'while target_qps_per_replica is not set. To enable '
                        'autoscaling, please set target_qps_per_replica.')

        if not readiness_path.startswith('/'):
            with ux_utils.print_exception_no_traceback():
                raise ValueError('readiness_path must start with a slash (/). '
                                 f'Got: {readiness_path}')

        # Add the check for unknown load balancing policies
        if (load_balancing_policy is not None and
                load_balancing_policy not in serve.LB_POLICIES):
            with ux_utils.print_exception_no_traceback():
                raise ValueError(
                    f'Unknown load balancing policy: {load_balancing_policy}. '
                    f'Available policies: {list(serve.LB_POLICIES.keys())}')
        self._readiness_path: str = readiness_path
        self._initial_delay_seconds: int = initial_delay_seconds
        self._readiness_timeout_seconds: int = readiness_timeout_seconds
        self._min_replicas: int = min_replicas
        self._max_replicas: Optional[int] = max_replicas
        self._num_overprovision: Optional[int] = num_overprovision
        self._ports: Optional[str] = ports
        self._target_qps_per_replica: Optional[Union[float, Dict[
            str, float]]] = target_qps_per_replica
        self._post_data: Optional[Dict[str, Any]] = post_data
        self._tls_credential: Optional[serve_utils.TLSCredential] = (
            tls_credential)
        self._readiness_headers: Optional[Dict[str, str]] = readiness_headers
        self._dynamic_ondemand_fallback: Optional[
            bool] = dynamic_ondemand_fallback
        self._base_ondemand_fallback_replicas: Optional[
            int] = base_ondemand_fallback_replicas
        self._spot_placer: Optional[str] = spot_placer
        self._upscale_delay_seconds: Optional[int] = upscale_delay_seconds
        self._downscale_delay_seconds: Optional[int] = downscale_delay_seconds
        self._load_balancing_policy: Optional[str] = load_balancing_policy
        self._pool: Optional[bool] = pool

        self._use_ondemand_fallback: bool = (
            self.dynamic_ondemand_fallback is not None and
            self.dynamic_ondemand_fallback) or (
                self.base_ondemand_fallback_replicas is not None and
                self.base_ondemand_fallback_replicas > 0)

    @staticmethod
    def from_yaml_config(config: Dict[str, Any]) -> 'SkyServiceSpec':
        common_utils.validate_schema(config, schemas.get_service_schema(),
                                     'Invalid service YAML: ')
        if 'replicas' in config and 'replica_policy' in config:
            with ux_utils.print_exception_no_traceback():
                raise ValueError(
                    'Cannot specify both `replicas` and `replica_policy` in '
                    'the service YAML. Please use one of them.')

        service_config: Dict[str, Any] = {}

        readiness_section = config.get('readiness_probe', '/')
        if isinstance(readiness_section, str):
            service_config['readiness_path'] = readiness_section
            initial_delay_seconds = None
            post_data = None
            readiness_timeout_seconds = None
            readiness_headers = None
        else:
            service_config['readiness_path'] = readiness_section['path']
            initial_delay_seconds = readiness_section.get(
                'initial_delay_seconds', None)
            post_data = readiness_section.get('post_data', None)
            readiness_timeout_seconds = readiness_section.get(
                'timeout_seconds', None)
            readiness_headers = readiness_section.get('headers', None)
        if initial_delay_seconds is None:
            initial_delay_seconds = constants.DEFAULT_INITIAL_DELAY_SECONDS
        service_config['initial_delay_seconds'] = initial_delay_seconds
        if readiness_timeout_seconds is None:
            readiness_timeout_seconds = (
                constants.DEFAULT_READINESS_PROBE_TIMEOUT_SECONDS)
        service_config['readiness_timeout_seconds'] = readiness_timeout_seconds
        if isinstance(post_data, str):
            try:
                post_data = json.loads(post_data)
            except json.JSONDecodeError as e:
                with ux_utils.print_exception_no_traceback():
                    raise ValueError(
                        'Invalid JSON string for `post_data` in the '
                        '`readiness_probe` section of your service YAML.'
                    ) from e
        service_config['post_data'] = post_data
        service_config['readiness_headers'] = readiness_headers

        ports = config.get('ports', None)
        if ports is not None:
            assert isinstance(ports, int)
            if not 1 <= ports <= 65535:
                with ux_utils.print_exception_no_traceback():
                    raise ValueError('Port must be between 1 and 65535.')
        service_config['ports'] = str(ports) if ports is not None else None

        pool_config = config.get('pool', None)
        if pool_config is not None:
            service_config['pool'] = pool_config

        policy_section = config.get('replica_policy', None)
        if policy_section is not None and pool_config:
            with ux_utils.print_exception_no_traceback():
                raise ValueError('Cannot specify `replica_policy` for cluster '
                                 'pool. Only `workers: <num>` is supported '
                                 'for cluster pool now.')

        simplified_policy_section = config.get('replicas', None)
        workers_config = config.get('workers', None)
        if simplified_policy_section is not None and workers_config is not None:
            with ux_utils.print_exception_no_traceback():
                raise ValueError('Cannot specify both `replicas` and `workers`.'
                                 ' Please use one of them.')
        if simplified_policy_section is not None and pool_config:
            with ux_utils.print_exception_no_traceback():
                raise ValueError('Cannot specify `replicas` for cluster pool. '
                                 'Please use `workers` instead.')
        if simplified_policy_section is None:
            simplified_policy_section = workers_config
        if policy_section is None or simplified_policy_section is not None:
            if simplified_policy_section is not None:
                min_replicas = simplified_policy_section
            else:
                min_replicas = constants.DEFAULT_MIN_REPLICAS
            service_config['min_replicas'] = min_replicas
            service_config['max_replicas'] = None
            service_config['num_overprovision'] = None
            service_config['target_qps_per_replica'] = None
            service_config['upscale_delay_seconds'] = None
            service_config['downscale_delay_seconds'] = None
        else:
            service_config['min_replicas'] = policy_section['min_replicas']
            service_config['max_replicas'] = policy_section.get(
                'max_replicas', None)
            service_config['num_overprovision'] = policy_section.get(
                'num_overprovision', None)
            service_config['target_qps_per_replica'] = policy_section.get(
                'target_qps_per_replica', None)
            service_config['upscale_delay_seconds'] = policy_section.get(
                'upscale_delay_seconds', None)
            service_config['downscale_delay_seconds'] = policy_section.get(
                'downscale_delay_seconds', None)
            service_config[
                'base_ondemand_fallback_replicas'] = policy_section.get(
                    'base_ondemand_fallback_replicas', None)
            service_config['dynamic_ondemand_fallback'] = policy_section.get(
                'dynamic_ondemand_fallback', None)
            service_config['spot_placer'] = policy_section.get(
                'spot_placer', None)

        service_config['load_balancing_policy'] = config.get(
            'load_balancing_policy', None)

        # Validate instance-aware settings
        target_qps_per_replica = service_config['target_qps_per_replica']
        load_balancing_policy = service_config['load_balancing_policy']

        if isinstance(target_qps_per_replica, dict):
            if load_balancing_policy != 'instance_aware_least_load':
                with ux_utils.print_exception_no_traceback():
                    raise ValueError(
                        'When using dict type target_qps_per_replica, '
                        'load_balancing_policy must be '
                        '"instance_aware_least_load".')

        if load_balancing_policy == 'instance_aware_least_load':
            if not isinstance(target_qps_per_replica, dict):
                with ux_utils.print_exception_no_traceback():
                    raise ValueError(
                        'When using "instance_aware_least_load" policy, '
                        'target_qps_per_replica must be a '
                        'dict mapping GPU types to QPS values.')

        tls_section = config.get('tls', None)
        if tls_section is not None:
            service_config['tls_credential'] = serve_utils.TLSCredential(
                keyfile=tls_section.get('keyfile', None),
                certfile=tls_section.get('certfile', None),
            )

        return SkyServiceSpec(**service_config)

    @staticmethod
    def from_yaml(yaml_path: str) -> 'SkyServiceSpec':
        with open(os.path.expanduser(yaml_path), 'r', encoding='utf-8') as f:
            config = yaml_utils.safe_load(f)

        if isinstance(config, str):
            with ux_utils.print_exception_no_traceback():
                raise ValueError('YAML loaded as str, not as dict. '
                                 f'Is it correct? Path: {yaml_path}')

        if config is None:
            config = {}

        if 'service' not in config:
            with ux_utils.print_exception_no_traceback():
                raise ValueError('Service YAML must have a "service" section. '
                                 f'Is it correct? Path: {yaml_path}')

        return SkyServiceSpec.from_yaml_config(config['service'])

    def to_yaml_config(self) -> Dict[str, Any]:
        config: Dict[str, Any] = {}

        def add_if_not_none(section: str,
                            key: Optional[str],
                            value: Any,
                            no_empty: bool = False):
            if no_empty and not value:
                return
            if value is not None:
                if key is None:
                    config[section] = value
                else:
                    if section not in config:
                        config[section] = dict()
                    config[section][key] = value

        add_if_not_none('pool', None, self._pool)

        if self.pool:
            # For pool, currently only `workers: <num>` is supported.
            add_if_not_none('workers', None, self.min_replicas)
            return config

        add_if_not_none('readiness_probe', 'path', self.readiness_path)
        add_if_not_none('readiness_probe', 'initial_delay_seconds',
                        self.initial_delay_seconds)
        add_if_not_none('readiness_probe', 'post_data', self.post_data)
        add_if_not_none('readiness_probe', 'timeout_seconds',
                        self.readiness_timeout_seconds)
        add_if_not_none('readiness_probe', 'headers', self._readiness_headers)
        add_if_not_none('replica_policy', 'min_replicas', self.min_replicas)
        add_if_not_none('replica_policy', 'max_replicas', self.max_replicas)
        add_if_not_none('replica_policy', 'num_overprovision',
                        self.num_overprovision)
        add_if_not_none('replica_policy', 'target_qps_per_replica',
                        self.target_qps_per_replica)
        add_if_not_none('replica_policy', 'dynamic_ondemand_fallback',
                        self.dynamic_ondemand_fallback)
        add_if_not_none('replica_policy', 'base_ondemand_fallback_replicas',
                        self.base_ondemand_fallback_replicas)
        add_if_not_none('replica_policy', 'spot_placer', self.spot_placer)
        add_if_not_none('replica_policy', 'upscale_delay_seconds',
                        self.upscale_delay_seconds)
        add_if_not_none('replica_policy', 'downscale_delay_seconds',
                        self.downscale_delay_seconds)
        add_if_not_none('load_balancing_policy', None,
                        self._load_balancing_policy)
        add_if_not_none('ports', None, int(self.ports) if self.ports else None)
        if self.tls_credential is not None:
            add_if_not_none('tls', 'keyfile', self.tls_credential.keyfile)
            add_if_not_none('tls', 'certfile', self.tls_credential.certfile)
        return config

    def probe_str(self):
        if self.post_data is None:
            method = f'GET {self.readiness_path}'
        else:
            method = f'POST {self.readiness_path} {json.dumps(self.post_data)}'
        headers = ('' if self.readiness_headers is None else
                   ' with custom headers')
        return f'{method}{headers}'

    def spot_policy_str(self) -> str:
        policy_strs: List[str] = []
        if (self.dynamic_ondemand_fallback is not None and
                self.dynamic_ondemand_fallback):
            if self.spot_placer is not None:
                if self.spot_placer == spot_placer_lib.SPOT_HEDGE_PLACER:
                    return 'SpotHedge'
            policy_strs.append('Dynamic on-demand fallback')
            if self.base_ondemand_fallback_replicas is not None:
                policy_strs.append(
                    f'with {self.base_ondemand_fallback_replicas}'
                    'base on-demand replicas')
        else:
            if self.base_ondemand_fallback_replicas is not None:
                plural = (''
                          if self.base_ondemand_fallback_replicas == 1 else 's')
                policy_strs.append('Static spot mixture with '
                                   f'{self.base_ondemand_fallback_replicas} '
                                   f'base on-demand replica{plural}')
        if self.spot_placer is not None:
            if not policy_strs:
                policy_strs.append('Spot placement')
            policy_strs.append(f'with {self.spot_placer} placer')
        if not policy_strs:
            return 'No spot policy'
        return ' '.join(policy_strs)

    def autoscaling_policy_str(self):
        if self.pool:
            # We only support fixed-size pool for now.
            return f'Fixed-size ({self.min_replicas} workers)'
        # TODO(MaoZiming): Update policy_str
        noun = 'worker' if self.pool else 'replica'
        min_plural = '' if self.min_replicas == 1 else 's'
        if self.max_replicas == self.min_replicas or self.max_replicas is None:
            return f'Fixed {self.min_replicas} {noun}{min_plural}'
        # Already checked in __init__.
        assert self.target_qps_per_replica is not None
        # TODO(tian): Refactor to contain more information
        max_plural = '' if self.max_replicas == 1 else 's'
        overprovision_str = ''
        if self.num_overprovision is not None:
            overprovision_str = (
                f' with {self.num_overprovision} overprovisioned replicas')
        return (f'Autoscaling from {self.min_replicas} to {self.max_replicas} '
                f'{noun}{max_plural}{overprovision_str} (target QPS per '
                f'{noun}: {self.target_qps_per_replica})')

    def set_ports(self, ports: str) -> None:
        self._ports = ports

    def tls_str(self):
        if self.tls_credential is None:
            return 'No TLS Enabled'
        return (f'Keyfile: {self.tls_credential.keyfile}, '
                f'Certfile: {self.tls_credential.certfile}')

    def __repr__(self) -> str:
        if self.pool:
            return textwrap.dedent(f"""\
                Worker policy:  {self.autoscaling_policy_str()}
            """)
        return textwrap.dedent(f"""\
            Readiness probe method:           {self.probe_str()}
            Readiness initial delay seconds:  {self.initial_delay_seconds}
            Readiness probe timeout seconds:  {self.readiness_timeout_seconds}
            Replica autoscaling policy:       {self.autoscaling_policy_str()}
            TLS Certificates:                 {self.tls_str()}
            Spot Policy:                      {self.spot_policy_str()}
            Load Balancing Policy:            {self.load_balancing_policy}
        """)

    @property
    def readiness_path(self) -> str:
        return self._readiness_path

    @property
    def initial_delay_seconds(self) -> int:
        return self._initial_delay_seconds

    @property
    def readiness_timeout_seconds(self) -> int:
        return self._readiness_timeout_seconds

    @property
    def min_replicas(self) -> int:
        return self._min_replicas

    @property
    def max_replicas(self) -> Optional[int]:
        # If None, treated as having the same value of min_replicas.
        return self._max_replicas

    @property
    def num_overprovision(self) -> Optional[int]:
        return self._num_overprovision

    @property
    def ports(self) -> Optional[str]:
        return self._ports

    @property
    def target_qps_per_replica(
            self) -> Optional[Union[float, Dict[str, float]]]:
        return self._target_qps_per_replica

    @property
    def post_data(self) -> Optional[Dict[str, Any]]:
        return self._post_data

    @property
    def tls_credential(self) -> Optional[serve_utils.TLSCredential]:
        return self._tls_credential

    @tls_credential.setter
    def tls_credential(self,
                       value: Optional[serve_utils.TLSCredential]) -> None:
        self._tls_credential = value

    @property
    def readiness_headers(self) -> Optional[Dict[str, str]]:
        return self._readiness_headers

    @property
    def base_ondemand_fallback_replicas(self) -> Optional[int]:
        return self._base_ondemand_fallback_replicas

    @property
    def dynamic_ondemand_fallback(self) -> Optional[bool]:
        return self._dynamic_ondemand_fallback

    @property
    def spot_placer(self) -> Optional[str]:
        return self._spot_placer

    @property
    def upscale_delay_seconds(self) -> Optional[int]:
        return self._upscale_delay_seconds

    @property
    def downscale_delay_seconds(self) -> Optional[int]:
        return self._downscale_delay_seconds

    @property
    def use_ondemand_fallback(self) -> bool:
        return self._use_ondemand_fallback

    @property
    def load_balancing_policy(self) -> str:
        return lb_policies.LoadBalancingPolicy.make_policy_name(
            self._load_balancing_policy)

    @property
    def pool(self) -> bool:
        # This can happen for backward compatibility.
        if not hasattr(self, '_pool'):
            return False
        return bool(self._pool)



"""Kubernetes."""
import os
import re
import subprocess
import tempfile
from typing import Dict, Iterator, List, Optional, Set, Tuple, Union

import colorama

from sky import catalog
from sky import clouds
from sky import exceptions
from sky import resources as resources_lib
from sky import sky_logging
from sky import skypilot_config
from sky.adaptors import kubernetes
from sky.clouds.utils import gcp_utils
from sky.provision import instance_setup
from sky.provision.gcp import constants as gcp_constants
from sky.provision.kubernetes import network_utils
from sky.provision.kubernetes import utils as kubernetes_utils
from sky.provision.kubernetes.utils import is_tpu_on_gke
from sky.provision.kubernetes.utils import KubernetesHighPerformanceNetworkType
from sky.provision.kubernetes.utils import normalize_tpu_accelerator_name
from sky.skylet import constants
from sky.utils import annotations
from sky.utils import common_utils
from sky.utils import kubernetes_enums
from sky.utils import registry
from sky.utils import resources_utils
from sky.utils import schemas
from sky.utils import volume as volume_lib

logger = sky_logging.init_logger(__name__)

# Namespace for SkyPilot resources shared across multiple tenants on the
# same cluster (even if they might be running in different namespaces).
# E.g., FUSE device manager daemonset is run in this namespace.
_SKYPILOT_SYSTEM_NAMESPACE = 'skypilot-system'

# Shared directory to communicate with fusermount-server, refer to
# addons/fuse-proxy/README.md for more details.
_FUSERMOUNT_SHARED_DIR = '/var/run/fusermount'


@registry.CLOUD_REGISTRY.register(aliases=['k8s'])
class Kubernetes(clouds.Cloud):
    """Kubernetes."""

    SKY_SSH_KEY_SECRET_NAME = 'sky-ssh-keys'
    SKY_SSH_JUMP_NAME = 'sky-ssh-jump-pod'

    # Limit the length of the cluster name to avoid exceeding the limit of 63
    # characters for Kubernetes resources. We limit to 42 characters (63-21) to
    # allow additional characters for creating ingress services to expose ports.
    # These services are named as {cluster_name_on_cloud}--skypilot-svc--{port},
    # where the suffix is 21 characters long.
    _MAX_CLUSTER_NAME_LEN_LIMIT = 42

    _MAX_VOLUME_NAME_LEN_LIMIT = 253

    _SUPPORTS_SERVICE_ACCOUNT_ON_REMOTE = True

    _DEFAULT_NUM_VCPUS = 2
    _DEFAULT_MEMORY_CPU_RATIO = 1
    _DEFAULT_MEMORY_CPU_RATIO_WITH_GPU = 4  # Allocate more memory for GPU tasks
    _REPR = 'Kubernetes'
    _CLOUD_UNSUPPORTED_FEATURES = {
        # TODO(romilb): Stopping might be possible to implement with
        #  container checkpointing introduced in Kubernetes v1.25. See:
        #  https://kubernetes.io/blog/2022/12/05/forensic-container-checkpointing-alpha/ # pylint: disable=line-too-long
        clouds.CloudImplementationFeatures.STOP: 'Kubernetes does not '
                                                 'support stopping VMs.',
        clouds.CloudImplementationFeatures.SPOT_INSTANCE: 'Spot instances are '
                                                          'not supported in '
                                                          'Kubernetes.',
        clouds.CloudImplementationFeatures.CUSTOM_DISK_TIER: 'Custom disk '
                                                             'tiers are not '
                                                             'supported in '
                                                             'Kubernetes.',
        clouds.CloudImplementationFeatures.CUSTOM_MULTI_NETWORK:
            ('Customized multiple network interfaces are not supported in '
             'Kubernetes.'),
        clouds.CloudImplementationFeatures.CUSTOM_NETWORK_TIER:
            ('Custom network tier is not supported in this Kubernetes '
             'cluster.'),
    }

    IMAGE_CPU = 'skypilot:custom-cpu-ubuntu-2004'
    IMAGE_GPU = 'skypilot:custom-gpu-ubuntu-2004'

    PROVISIONER_VERSION = clouds.ProvisionerVersion.SKYPILOT
    STATUS_VERSION = clouds.StatusVersion.SKYPILOT

    _INDENT_PREFIX = ' ' * 4

    # Set of contexts that has logged as temporarily unreachable
    logged_unreachable_contexts: Set[str] = set()

    @property
    def ssh_key_secret_field_name(self):
        # Use a fresh user hash to avoid conflicts in the secret object naming.
        # This can happen when the controller is reusing the same user hash
        # through USER_ID_ENV_VAR but has a different SSH key.
        fresh_user_hash = common_utils.generate_user_hash()
        return f'ssh-publickey-{fresh_user_hash}'

    @classmethod
    def _unsupported_features_for_resources(
        cls, resources: 'resources_lib.Resources'
    ) -> Dict[clouds.CloudImplementationFeatures, str]:
        # TODO(aylei): features need to be regional (per context) to make
        # multi-kubernetes selection/failover work.
        unsupported_features = cls._CLOUD_UNSUPPORTED_FEATURES.copy()
        context = resources.region
        if context is None:
            context = kubernetes_utils.get_current_kube_config_context_name()
        unsupported_features[clouds.CloudImplementationFeatures.STOP] = (
            'Stopping clusters is not supported on Kubernetes.')
        unsupported_features[clouds.CloudImplementationFeatures.AUTOSTOP] = (
            'Auto-stop is not supported on Kubernetes.')
        # Allow spot instances if supported by the cluster
        try:
            spot_label_key, _ = kubernetes_utils.get_spot_label(context)
            if spot_label_key is not None:
                unsupported_features.pop(
                    clouds.CloudImplementationFeatures.SPOT_INSTANCE, None)
            # Allow custom network tier if supported by the cluster
            # (e.g., Nebius clusters with high performance networking)
            network_type, _ = cls._detect_network_type(context,
                                                       resources.network_tier)
            if network_type.supports_high_performance_networking():
                unsupported_features.pop(
                    clouds.CloudImplementationFeatures.CUSTOM_NETWORK_TIER,
                    None)
        except exceptions.KubeAPIUnreachableError as e:
            cls._log_unreachable_context(context, str(e))
        return unsupported_features

    @classmethod
    def max_cluster_name_length(cls) -> Optional[int]:
        return cls._MAX_CLUSTER_NAME_LEN_LIMIT

    @classmethod
    @annotations.lru_cache(scope='global', maxsize=1)
    def _log_skipped_contexts_once(cls, skipped_contexts: Tuple[str,
                                                                ...]) -> None:
        """Log skipped contexts for only once.

        We don't directly cache the result of _filter_existing_allowed_contexts
        as the admin policy may update the allowed contexts.
        """
        if skipped_contexts:
            logger.warning(
                f'Kubernetes contexts {set(skipped_contexts)!r} specified in '
                '"allowed_contexts" not found in kubeconfig. '
                'Ignoring these contexts.')

    @classmethod
    def existing_allowed_contexts(cls, silent: bool = False) -> List[str]:
        """Get existing allowed contexts.

        If None is returned in the list, it means that we are running in a pod
        with in-cluster auth. In this case, we specify None context, which will
        use the service account mounted in the pod.
        """
        all_contexts = kubernetes_utils.get_all_kube_context_names()
        if not all_contexts:
            return []

        all_contexts = set(all_contexts)

        # Allowed_contexts specified for workspace should take precedence over
        # the global allowed_contexts.
        allowed_contexts = skypilot_config.get_workspace_cloud(
            'kubernetes').get('allowed_contexts', None)
        if allowed_contexts is None:
            allowed_contexts = skypilot_config.get_effective_region_config(
                cloud='kubernetes',
                region=None,
                keys=('allowed_contexts',),
                default_value=None)

        # Exclude contexts starting with `ssh-`
        # TODO(romilb): Remove when SSH Node Pools use a separate kubeconfig.
        all_contexts = [
            ctx for ctx in all_contexts if not ctx.startswith('ssh-')
        ]

        if allowed_contexts is None:
            # Try kubeconfig if present
            current_context = (
                kubernetes_utils.get_current_kube_config_context_name())
            if ((current_context is None or current_context.startswith('ssh-'))
                    and kubernetes_utils.is_incluster_config_available()):
                # If no kubeconfig contexts found, use in-cluster if available
                current_context = kubernetes.in_cluster_context_name()
            allowed_contexts = []
            if current_context is not None:
                allowed_contexts = [current_context]

        existing_contexts = []
        skipped_contexts = []
        for context in allowed_contexts:
            if context in all_contexts:
                existing_contexts.append(context)
            else:
                # Skip SSH Node Pool contexts
                if context.startswith('ssh-'):
                    continue
                skipped_contexts.append(context)
        if not silent:
            cls._log_skipped_contexts_once(tuple(skipped_contexts))
        return existing_contexts

    @classmethod
    def _log_unreachable_context(cls,
                                 context: str,
                                 reason: Optional[str] = None) -> None:
        """Logs a Kubernetes context as unreachable.

        Args:
            context: The Kubernetes context to mark as unreachable.
            reason: Optional reason for marking the context as unreachable.
            silent: Whether to suppress the log message.
        """
        # Skip if this context has already been logged as unreachable
        if context in cls.logged_unreachable_contexts:
            return

        cls.logged_unreachable_contexts.add(context)
        msg = f'Excluding Kubernetes context {context}'
        if reason is not None:
            msg += f': {reason}'
        logger.info(msg)

        # Check if all existing allowed contexts are now unreachable
        existing_contexts = cls.existing_allowed_contexts()
        if existing_contexts and all(ctx in cls.logged_unreachable_contexts
                                     for ctx in existing_contexts):
            logger.warning(
                'All Kubernetes contexts are unreachable. '
                'Retry if it is a transient error, or run sky check to '
                'refresh Kubernetes availability if permanent.')

    @classmethod
    def regions_with_offering(cls, instance_type: Optional[str],
                              accelerators: Optional[Dict[str, int]],
                              use_spot: bool, region: Optional[str],
                              zone: Optional[str]) -> List[clouds.Region]:
        del accelerators, zone, use_spot  # unused
        existing_contexts = cls.existing_allowed_contexts()

        regions = []
        for context in existing_contexts:
            regions.append(clouds.Region(context))

        if region is not None:
            regions = [r for r in regions if r.name == region]

        # Check if requested instance type will fit in the cluster.
        # TODO(zhwu,romilb): autoscaler type needs to be regional (per
        # kubernetes cluster/context).
        if instance_type is None:
            return regions

        regions_to_return = []
        for r in regions:
            context = r.name
            try:
                fits, reason = kubernetes_utils.check_instance_fits(
                    context, instance_type)
            except exceptions.KubeAPIUnreachableError as e:
                cls._log_unreachable_context(context, str(e))
                continue
            if fits:
                regions_to_return.append(r)
                continue
            logger.debug(f'Instance type {instance_type} does '
                         'not fit in the existing Kubernetes cluster '
                         'with context: '
                         f'{context}. Reason: {reason}')

            autoscaler_type = skypilot_config.get_effective_region_config(
                cloud='kubernetes',
                region=context,
                keys=('autoscaler',),
                default_value=None)
            if (autoscaler_type is not None and
                    not kubernetes_utils.get_autoscaler(
                        kubernetes_enums.KubernetesAutoscalerType(
                            autoscaler_type)).can_query_backend):
                # Unsupported autoscaler type. Rely on the autoscaler to
                # provision the right instance type without running checks.
                # Worst case, if autoscaling fails, the pod will be stuck in
                # pending state until provision_timeout, after which failover
                # will be triggered.
                #
                # Removing this if statement produces the same behavior,
                # because can_create_new_instance_of_type() always returns True
                # for unsupported autoscaler types.
                # This check is here as a performance optimization to avoid
                # further code executions that is known to return this result.
                regions_to_return.append(r)
                continue

            if autoscaler_type is None:
                continue
            autoscaler = kubernetes_utils.get_autoscaler(
                kubernetes_enums.KubernetesAutoscalerType(autoscaler_type))
            logger.debug(f'{context} has autoscaler of type: {autoscaler_type}')
            if autoscaler.can_create_new_instance_of_type(
                    context, instance_type):
                logger.debug(f'Kubernetes cluster {context} can be '
                             'autoscaled to create instance type '
                             f'{instance_type}. Including {context} '
                             'in the list of regions to return.')
                regions_to_return.append(r)
        return regions_to_return

    def instance_type_to_hourly_cost(self,
                                     instance_type: str,
                                     use_spot: bool,
                                     region: Optional[str] = None,
                                     zone: Optional[str] = None) -> float:
        # TODO(romilb): Investigate how users can provide their own cost catalog
        #  for Kubernetes clusters.
        # For now, assume zero cost for Kubernetes clusters
        return 0.0

    def accelerators_to_hourly_cost(self,
                                    accelerators: Dict[str, int],
                                    use_spot: bool,
                                    region: Optional[str] = None,
                                    zone: Optional[str] = None) -> float:
        del accelerators, use_spot, region, zone  # unused
        return 0.0

    def get_egress_cost(self, num_gigabytes: float) -> float:
        return 0.0

    def __repr__(self):
        return self._REPR

    @classmethod
    def get_default_instance_type(
            cls,
            cpus: Optional[str] = None,
            memory: Optional[str] = None,
            disk_tier: Optional['resources_utils.DiskTier'] = None,
            region: Optional[str] = None,
            zone: Optional[str] = None) -> str:
        # TODO(romilb): In the future, we may want to move the instance type
        #  selection + availability checking to a kubernetes_catalog module.
        del disk_tier, region, zone  # Unused.
        # We strip '+' from resource requests since Kubernetes can provision
        # exactly the requested resources.
        instance_cpus = float(
            cpus.strip('+')) if cpus is not None else cls._DEFAULT_NUM_VCPUS
        if memory is not None:
            if memory.endswith('+'):
                instance_mem = float(memory[:-1])
            elif memory.endswith('x'):
                instance_mem = float(memory[:-1]) * instance_cpus
            else:
                instance_mem = float(memory)
        else:
            instance_mem = instance_cpus * cls._DEFAULT_MEMORY_CPU_RATIO
        virtual_instance_type = kubernetes_utils.KubernetesInstanceType(
            instance_cpus, instance_mem).name
        return virtual_instance_type

    @classmethod
    def get_accelerators_from_instance_type(
        cls,
        instance_type: str,
    ) -> Optional[Dict[str, Union[int, float]]]:
        inst = kubernetes_utils.KubernetesInstanceType.from_instance_type(
            instance_type)
        return {
            inst.accelerator_type: inst.accelerator_count
        } if (inst.accelerator_count is not None and
              inst.accelerator_type is not None) else None

    @classmethod
    def get_vcpus_mem_from_instance_type(
            cls, instance_type: str) -> Tuple[Optional[float], Optional[float]]:
        """Returns the #vCPUs and memory that the instance type offers."""
        k = kubernetes_utils.KubernetesInstanceType.from_instance_type(
            instance_type)
        return k.cpus, k.memory

    @classmethod
    def zones_provision_loop(
        cls,
        *,
        region: str,
        num_nodes: int,
        instance_type: str,
        accelerators: Optional[Dict[str, int]] = None,
        use_spot: bool = False,
    ) -> Iterator[Optional[List[clouds.Zone]]]:
        # Always yield None for zones, since Kubernetes does not have zones, and
        # we should allow any region get to this point.
        yield None

    @classmethod
    def get_zone_shell_cmd(cls) -> Optional[str]:
        return None

    @classmethod
    def get_image_size(cls, image_id: str, region: Optional[str]) -> int:
        del image_id, region  # Unused.
        # We don't limit the image by its size compared to the disk size, as
        # we don't have a notion of disk size in Kubernetes.
        return 0

    @staticmethod
    def _calculate_provision_timeout(
        num_nodes: int,
        volume_mounts: Optional[List['volume_lib.VolumeMount']],
        enable_flex_start: bool,
    ) -> int:
        """Calculate provision timeout based on number of nodes.

        The timeout scales linearly with the number of nodes to account for
        scheduling overhead, but is capped to avoid excessive waiting.

        Args:
            num_nodes: Number of nodes being provisioned
            volume_mounts: Volume mounts for the pod
            enable_flex_start: Whether flex start is enabled

        Returns:
            Timeout in seconds
        """
        base_timeout = 10  # Base timeout for single node
        per_node_timeout = 0.2  # Additional seconds per node
        max_timeout = 60  # Cap at 1 minute
        if enable_flex_start:
            # Flex start takes longer to provision.
            base_timeout = 1200
            per_node_timeout = 10
            max_timeout = 2400
        elif volume_mounts is not None:
            for volume_mount in volume_mounts:
                if (volume_mount.volume_config.type ==
                        volume_lib.VolumeType.PVC.value):
                    if (volume_mount.volume_config.config.get(
                            'access_mode', '') ==
                            volume_lib.VolumeAccessMode.READ_WRITE_MANY.value):
                        # GKE may take several minutes to provision a PV
                        # supporting READ_WRITE_MANY with filestore.
                        base_timeout = 180
                        max_timeout = 240
                        break

        return int(
            min(base_timeout + (per_node_timeout * (num_nodes - 1)),
                max_timeout))

    def make_deploy_resources_variables(
        self,
        resources: 'resources_lib.Resources',
        cluster_name: 'resources_utils.ClusterName',
        region: Optional['clouds.Region'],
        zones: Optional[List['clouds.Zone']],
        num_nodes: int,
        dryrun: bool = False,
        volume_mounts: Optional[List['volume_lib.VolumeMount']] = None,
    ) -> Dict[str, Optional[str]]:
        del cluster_name, zones, dryrun  # Unused.
        if region is None:
            context = kubernetes_utils.get_current_kube_config_context_name()
        else:
            context = region.name
        assert context is not None, 'No context found in kubeconfig'

        resources = resources.assert_launchable()
        acc_dict = self.get_accelerators_from_instance_type(
            resources.instance_type)
        custom_resources = resources_utils.make_ray_custom_resources_str(
            acc_dict)

        # resources.memory and cpus are None if they are not explicitly set.
        # We fetch the default values for the instance type in that case.
        k = kubernetes_utils.KubernetesInstanceType.from_instance_type(
            resources.instance_type)
        cpus = k.cpus
        mem = k.memory
        # Optionally populate accelerator information.
        acc_type = k.accelerator_type
        acc_count = k.accelerator_count
        if acc_type is not None and is_tpu_on_gke(acc_type):
            acc_type, acc_count = normalize_tpu_accelerator_name(acc_type)
        else:
            acc_count = acc_count or 0

        def _get_image_id(resources: 'resources_lib.Resources') -> str:
            image_id_dict = resources.image_id
            if image_id_dict is not None:
                # Use custom image specified in resources
                if None in image_id_dict:
                    image_id = image_id_dict[None]
                else:
                    assert resources.region in image_id_dict, image_id_dict
                    image_id = image_id_dict[resources.region]
                if image_id.startswith('docker:'):
                    image_id = image_id[len('docker:'):]
            else:
                # Select image based on whether we are using GPUs or not.
                image_id = self.IMAGE_GPU if acc_count > 0 else self.IMAGE_CPU
                # Get the container image ID from the service catalog.
                image_id = catalog.get_image_id_from_tag(image_id,
                                                         clouds='kubernetes')
            return image_id

        image_id = _get_image_id(resources)
        # TODO(romilb): Create a lightweight image for SSH jump host
        ssh_jump_image = catalog.get_image_id_from_tag(self.IMAGE_CPU,
                                                       clouds='kubernetes')

        # Set environment variables for the pod. Note that SkyPilot env vars
        # are set separately when the task is run. These env vars are
        # independent of the SkyPilot task to be run.
        k8s_env_vars = {kubernetes.IN_CLUSTER_CONTEXT_NAME_ENV_VAR: context}

        # Setup GPU/TPU labels and resource keys.
        k8s_acc_label_key = None
        k8s_acc_label_values = None
        k8s_topology_label_key = None
        k8s_topology_label_value = None
        k8s_resource_key = None
        tpu_requested = False
        avoid_label_keys = None

        # If GPU/TPUs are requested, set node label to match the GPU/TPU type.
        if acc_count > 0 and acc_type is not None:
            (k8s_acc_label_key, k8s_acc_label_values, k8s_topology_label_key,
             k8s_topology_label_value) = (
                 kubernetes_utils.get_accelerator_label_key_values(
                     context, acc_type, acc_count))
            if (k8s_acc_label_key ==
                    kubernetes_utils.GKELabelFormatter.TPU_LABEL_KEY):
                tpu_requested = True
                k8s_resource_key = kubernetes_utils.TPU_RESOURCE_KEY
            else:
                k8s_resource_key = kubernetes_utils.get_gpu_resource_key(
                    context)
        else:
            # If no GPUs are requested, we set NVIDIA_VISIBLE_DEVICES=none to
            # maintain GPU isolation. This is to override the default behavior
            # of Nvidia device plugin which would expose all GPUs to the pod
            # when no GPUs are requested.
            # Note that NVIDIA_VISIBLE_DEVICES is different from
            # CUDA_VISIBLE_DEVICES - the latter is used to control which GPUs
            # are visible to the application and is set inside the pod, while
            # the former is used to control which GPUs are visible to the pod
            # through the nvidia runtime.
            # See: https://github.com/NVIDIA/k8s-device-plugin/issues/61
            k8s_env_vars['NVIDIA_VISIBLE_DEVICES'] = 'none'
            avoid_label_keys = kubernetes_utils.get_accelerator_label_keys(
                context)
            if len(avoid_label_keys) == 0:
                avoid_label_keys = None
        port_mode = network_utils.get_port_mode(None, context)

        remote_identity = skypilot_config.get_effective_region_config(
            cloud='kubernetes',
            region=context,
            keys=('remote_identity',),
            default_value=schemas.get_default_remote_identity('kubernetes'))

        if isinstance(remote_identity, dict):
            # If remote_identity is a dict, use the service account for the
            # current context
            k8s_service_account_name = remote_identity.get(context, None)
            if k8s_service_account_name is None:
                err_msg = (f'Context {context!r} not found in '
                           'remote identities from config.yaml')
                raise ValueError(err_msg)
        else:
            # If remote_identity is not a dict, use
            k8s_service_account_name = remote_identity

        lc = schemas.RemoteIdentityOptions.LOCAL_CREDENTIALS.value
        sa = schemas.RemoteIdentityOptions.SERVICE_ACCOUNT.value

        if k8s_service_account_name == lc or k8s_service_account_name == sa:
            # Use the default service account if remote identity is not set.
            # For LOCAL_CREDENTIALS, this is for in-cluster authentication
            # which needs a serviceaccount (specifically for SSH node pools
            # which uses in-cluster authentication internally, and we would
            # like to support exec-auth when the user is also using SSH infra)
            k8s_service_account_name = (
                kubernetes_utils.DEFAULT_SERVICE_ACCOUNT_NAME)

        fuse_device_required = bool(resources.requires_fuse)

        # Configure spot labels, if requested and supported
        spot_label_key, spot_label_value = None, None
        if resources.use_spot:
            spot_label_key, spot_label_value = kubernetes_utils.get_spot_label()

        network_type, machine_type = self._detect_network_type(
            context, resources.network_tier)

        # Check if this cluster supports high performance networking and
        # configure appropriate settings for different cluster types
        if (resources.network_tier is not None and
                resources.network_tier == resources_utils.NetworkTier.BEST):
            # Only proceed if CUSTOM_NETWORK_TIER is supported by this cluster
            unsupported_features = self._unsupported_features_for_resources(
                resources)
            if clouds.CloudImplementationFeatures.CUSTOM_NETWORK_TIER \
                    not in unsupported_features:
                # Add high-performance networking environment variables for
                # clusters with high performance networking
                network_env_vars = network_type.get_network_env_vars()
                k8s_env_vars.update(network_env_vars)

        # We specify object-store-memory to be 500MB to avoid taking up too
        # much memory on the head node. 'num-cpus' should be set to limit
        # the CPU usage on the head pod, otherwise the ray cluster will use the
        # CPU resources on the node instead within the pod.
        custom_ray_options = {
            'object-store-memory': 500000000,
            # 'num-cpus' must be an integer, but we should not set it to 0 if
            # cpus is <1.
            'num-cpus': str(max(int(cpus), 1)),
        }

        # Get the storage class name for high availability controller's PVC
        k8s_ha_storage_class_name = (
            skypilot_config.get_effective_region_config(
                cloud='kubernetes',
                region=context,
                keys=('high_availability', 'storage_class_name'),
                default_value=None))

        k8s_kueue_local_queue_name = (
            skypilot_config.get_effective_region_config(
                cloud='kubernetes',
                region=context,
                keys=('kueue', 'local_queue_name'),
                default_value=None,
                override_configs=resources.cluster_config_overrides))

        # Check DWS configuration for GKE.
        (enable_flex_start, enable_flex_start_queued_provisioning,
         max_run_duration_seconds) = gcp_utils.get_dws_config(
             context, k8s_kueue_local_queue_name,
             resources.cluster_config_overrides)
        if enable_flex_start_queued_provisioning or enable_flex_start:
            # DWS is only supported in GKE, check the autoscaler type.
            autoscaler_type = skypilot_config.get_effective_region_config(
                cloud='kubernetes',
                region=context,
                keys=('autoscaler',),
                default_value=None)
            if (autoscaler_type !=
                    kubernetes_enums.KubernetesAutoscalerType.GKE.value):
                raise ValueError(
                    f'DWS is only supported in GKE, but the autoscaler type '
                    f'for context {context} is {autoscaler_type}')

        # Timeout for resource provisioning. This timeout determines how long to
        # wait for pod to be in pending status before giving up.
        # Larger timeout may be required for autoscaling clusters, since
        # autoscaler may take some time to provision new nodes.
        # Note that this timeout includes time taken by the Kubernetes scheduler
        # itself, which can be upto 2-3 seconds, and up to 10-15 seconds when
        # scheduling 100s of pods.
        # We use a linear scaling formula to determine the timeout based on the
        # number of nodes.

        timeout = self._calculate_provision_timeout(
            num_nodes, volume_mounts, enable_flex_start or
            enable_flex_start_queued_provisioning)
        timeout = skypilot_config.get_effective_region_config(
            cloud='kubernetes',
            region=context,
            keys=('provision_timeout',),
            default_value=timeout,
            override_configs=resources.cluster_config_overrides)

        deploy_vars = {
            'instance_type': resources.instance_type,
            'custom_resources': custom_resources,
            'cpus': str(cpus),
            'memory': str(mem),
            'accelerator_count': str(acc_count),
            'timeout': str(timeout),
            'k8s_port_mode': port_mode.value,
            'k8s_networking_mode': network_utils.get_networking_mode(
                None, context=context).value,
            'k8s_ssh_key_secret_name': self.SKY_SSH_KEY_SECRET_NAME,
            'k8s_acc_label_key': k8s_acc_label_key,
            'k8s_acc_label_values': k8s_acc_label_values,
            'k8s_ssh_jump_name': self.SKY_SSH_JUMP_NAME,
            'k8s_ssh_jump_image': ssh_jump_image,
            'k8s_service_account_name': k8s_service_account_name,
            'k8s_automount_sa_token': 'true',
            'k8s_fuse_device_required': fuse_device_required,
            'k8s_kueue_local_queue_name': k8s_kueue_local_queue_name,
            # Namespace to run the fusermount-server daemonset in
            'k8s_skypilot_system_namespace': _SKYPILOT_SYSTEM_NAMESPACE,
            'k8s_fusermount_shared_dir': _FUSERMOUNT_SHARED_DIR,
            'k8s_spot_label_key': spot_label_key,
            'k8s_spot_label_value': spot_label_value,
            'tpu_requested': tpu_requested,
            'k8s_topology_label_key': k8s_topology_label_key,
            'k8s_topology_label_value': k8s_topology_label_value,
            'k8s_resource_key': k8s_resource_key,
            'k8s_env_vars': k8s_env_vars,
            'image_id': image_id,
            'ray_installation_commands': constants.RAY_INSTALLATION_COMMANDS,
            'ray_head_start_command': instance_setup.ray_head_start_command(
                custom_resources, custom_ray_options),
            'skypilot_ray_port': constants.SKY_REMOTE_RAY_PORT,
            'ray_worker_start_command': instance_setup.ray_worker_start_command(
                custom_resources, custom_ray_options, no_restart=False),
            'k8s_high_availability_deployment_volume_mount_name':
                (kubernetes_utils.HIGH_AVAILABILITY_DEPLOYMENT_VOLUME_MOUNT_NAME
                ),
            'k8s_high_availability_deployment_volume_mount_path':
                (kubernetes_utils.HIGH_AVAILABILITY_DEPLOYMENT_VOLUME_MOUNT_PATH
                ),
            'k8s_high_availability_deployment_setup_script_path':
                (constants.PERSISTENT_SETUP_SCRIPT_PATH),
            'k8s_high_availability_deployment_run_script_dir':
                (constants.PERSISTENT_RUN_SCRIPT_DIR),
            'k8s_high_availability_restarting_signal_file':
                (constants.PERSISTENT_RUN_RESTARTING_SIGNAL_FILE),
            'ha_recovery_log_path':
                constants.HA_PERSISTENT_RECOVERY_LOG_PATH.format(''),
            'sky_python_cmd': constants.SKY_PYTHON_CMD,
            'k8s_high_availability_storage_class_name':
                (k8s_ha_storage_class_name),
            'avoid_label_keys': avoid_label_keys,
            'k8s_enable_flex_start': enable_flex_start,
            'k8s_max_run_duration_seconds': max_run_duration_seconds,
            'k8s_network_type': network_type.value,
        }

        # Add kubecontext if it is set. It may be None if SkyPilot is running
        # inside a pod with in-cluster auth.
        if context is not None:
            deploy_vars['k8s_context'] = context

        namespace = kubernetes_utils.get_kube_config_context_namespace(context)
        deploy_vars['k8s_namespace'] = namespace

        # Add backward compatibility template variables for GPUDirect variants
        deploy_vars['k8s_enable_gpudirect_tcpx'] = (
            network_type == KubernetesHighPerformanceNetworkType.GCP_TCPX)
        deploy_vars['k8s_enable_gpudirect_tcpxo'] = (
            network_type == KubernetesHighPerformanceNetworkType.GCP_TCPXO)
        rdma_enabled = (network_type ==
                        KubernetesHighPerformanceNetworkType.GCP_GPUDIRECT_RDMA)
        deploy_vars['k8s_enable_gpudirect_rdma'] = rdma_enabled
        if rdma_enabled and machine_type.startswith('a4'):
            deploy_vars['k8s_enable_gpudirect_rdma_a4'] = True
        else:
            deploy_vars['k8s_enable_gpudirect_rdma_a4'] = False

        deploy_vars['k8s_ipc_lock_capability'] = (
            network_type.requires_ipc_lock_capability())

        return deploy_vars

    @staticmethod
    def _warn_on_disk_size(resources: 'resources_lib.Resources'):
        if resources.disk_size != resources_lib.DEFAULT_DISK_SIZE_GB:
            logger.info(f'{colorama.Style.DIM}Disk size {resources.disk_size} '
                        'is not supported by Kubernetes. '
                        'To add additional disk, use volumes.'
                        f'{colorama.Style.RESET_ALL}')
        if resources.disk_tier is not None:
            logger.info(f'{colorama.Style.DIM}Disk tier {resources.disk_tier} '
                        'is not supported by Kubernetes. '
                        'To add additional disk, use volumes.'
                        f'{colorama.Style.RESET_ALL}')

    def _get_feasible_launchable_resources(
        self, resources: 'resources_lib.Resources'
    ) -> 'resources_utils.FeasibleResources':
        # TODO(zhwu): This needs to be updated to return the correct region
        # (context) that has enough resources.
        self._warn_on_disk_size(resources)
        fuzzy_candidate_list: List[str] = []
        if resources.instance_type is not None:
            assert resources.is_launchable(), resources
            regions = self.regions_with_offering(
                resources.instance_type,
                accelerators=resources.accelerators,
                use_spot=resources.use_spot,
                region=resources.region,
                zone=resources.zone)
            if not regions:
                return resources_utils.FeasibleResources([], [], None)
            resources = resources.copy(accelerators=None)
            return resources_utils.FeasibleResources([resources],
                                                     fuzzy_candidate_list, None)

        def _make(instance_list):
            resource_list = []
            for instance_type in instance_list:
                r = resources.copy(
                    cloud=self.__class__(),
                    instance_type=instance_type,
                    accelerators=None,
                )
                resource_list.append(r)
            return resource_list

        # Currently, handle a filter on accelerators only.
        accelerators = resources.accelerators

        default_instance_type = Kubernetes.get_default_instance_type(
            cpus=resources.cpus,
            memory=resources.memory,
            disk_tier=resources.disk_tier,
            region=resources.region,
            zone=resources.zone)

        if accelerators is None:
            # For CPU only clusters, need no special handling
            chosen_instance_type = default_instance_type
        else:
            assert len(accelerators) == 1, resources
            # GPUs requested - build instance type.
            acc_type, acc_count = list(accelerators.items())[0]
            # If acc_type contains spaces, return empty list since Kubernetes
            # does not support spaces in label values
            if ' ' in acc_type:
                return resources_utils.FeasibleResources([], [], None)

            # Parse into KubernetesInstanceType
            k8s_instance_type = (kubernetes_utils.KubernetesInstanceType.
                                 from_instance_type(default_instance_type))

            gpu_task_cpus = k8s_instance_type.cpus
            # Special handling to bump up memory multiplier for GPU instances
            gpu_task_memory = (float(resources.memory.strip('+')) if
                               resources.memory is not None else gpu_task_cpus *
                               self._DEFAULT_MEMORY_CPU_RATIO_WITH_GPU)
            chosen_instance_type = (
                kubernetes_utils.KubernetesInstanceType.from_resources(
                    gpu_task_cpus, gpu_task_memory, acc_count, acc_type).name)
        # Check the availability of the specified instance type in all contexts.
        available_regions = self.regions_with_offering(
            chosen_instance_type,
            accelerators=None,
            use_spot=resources.use_spot,
            region=resources.region,
            zone=resources.zone)
        if not available_regions:
            return resources_utils.FeasibleResources([], [], None)
        # No fuzzy lists for Kubernetes
        # We don't set the resources returned with regions, because the
        # optimizer will further find the valid region (context) for the
        # resources.
        return resources_utils.FeasibleResources(_make([chosen_instance_type]),
                                                 [], None)

    @classmethod
    def _check_single_context(cls, context: str) -> Tuple[bool, str]:
        """Check if the user has access credentials to a single SSH context."""

        def _red_color(str_to_format: str) -> str:
            return (f'{colorama.Fore.LIGHTRED_EX}'
                    f'{str_to_format}'
                    f'{colorama.Style.RESET_ALL}')

        def _dim_color(str_to_format: str) -> str:
            return (f'{colorama.Style.DIM}'
                    f'{str_to_format}'
                    f'{colorama.Style.RESET_ALL}')

        def _bright_green_color(str_to_format: str) -> str:
            return (f'{colorama.Fore.GREEN}'
                    f'{str_to_format}'
                    f'{colorama.Style.RESET_ALL}')

        try:
            check_result = kubernetes_utils.check_credentials(
                context, run_optional_checks=True)
            if check_result[0]:
                if check_result[1] is not None:
                    return True, (_bright_green_color('enabled.') +
                                  _dim_color(f' Note: {check_result[1]}'))
                else:
                    return True, _bright_green_color('enabled.')
            else:
                assert check_result[1] is not None
                return False, (_red_color('disabled.') +
                               _dim_color(f' Reason: {check_result[1]}'))
        except Exception as e:  # pylint: disable=broad-except
            return False, _red_color(str(e))

    @classmethod
    def _check_compute_credentials(
            cls) -> Tuple[bool, Optional[Union[str, Dict[str, str]]]]:
        """Checks if the user has access credentials to
        Kubernetes."""
        # Check for port forward dependencies
        logger.debug(f'Checking compute credentials for {cls.canonical_name()}')
        reasons = kubernetes_utils.check_port_forward_mode_dependencies(False)
        if reasons is not None:
            formatted = '\n'.join(
                [reasons[0]] +
                [f'{cls._INDENT_PREFIX}' + r for r in reasons[1:]])
            return (False, formatted)

        # Test using python API
        try:
            existing_allowed_contexts = cls.existing_allowed_contexts()
        except ImportError as e:
            return (False,
                    f'{common_utils.format_exception(e, use_bracket=True)}')
        if not existing_allowed_contexts:
            if skypilot_config.loaded_config_path() is None:
                check_skypilot_config_msg = ''
            else:
                check_skypilot_config_msg = (
                    ' and check "allowed_contexts" in your '
                    f'{skypilot_config.loaded_config_path()} file.')
            return (False, 'No available context found in kubeconfig. '
                    'Check if you have a valid kubeconfig file' +
                    check_skypilot_config_msg)

        ctx2text = {}
        success = False
        for context in existing_allowed_contexts:
            suc, text = cls._check_single_context(context)
            success = success or suc
            ctx2text[context] = text

        return success, ctx2text

    @classmethod
    def _format_credential_check_results(cls, hints: List[str],
                                         reasons: List[str]) -> str:
        """Format credential check results with hints and reasons.

        Args:
            hints: List of successful context check messages.
            reasons: List of failed context check reasons.

        Returns:
            A formatted string containing hints and by failure reasons.
        """
        message_parts = []
        if len(hints) == 1 and not reasons:
            return hints[0]
        if hints:
            message_parts.append(f'\n{cls._INDENT_PREFIX}  ' +
                                 f'\n{cls._INDENT_PREFIX}  '.join(hints))
        if reasons:
            if hints:
                message_parts.append('\n')
            message_parts.append(
                f'\n{cls._INDENT_PREFIX}Unavailable contexts (remove from '
                '"allowed_contexts" config if permanently unavailable): '
                f'\n{cls._INDENT_PREFIX}  ' +
                f'\n{cls._INDENT_PREFIX}  '.join(reasons))
        return ''.join(message_parts)

    def get_credential_file_mounts(self) -> Dict[str, str]:
        credential_paths = kubernetes_utils.get_kubeconfig_paths()
        if credential_paths:
            # For single kubeconfig path, keep the original path.
            kubeconfig_file = credential_paths[0]
            if len(credential_paths) > 1:
                # For multiple kubeconfig paths, merge them into a single file.
                # TODO(aylei): GC merged kubeconfig files.
                kubeconfig_file = tempfile.NamedTemporaryFile(
                    prefix='merged-kubeconfig-', suffix='.yaml',
                    delete=False).name
                subprocess.run(
                    'kubectl config view --flatten '
                    f'> {kubeconfig_file}',
                    shell=True,
                    check=True)
            if os.path.exists(kubeconfig_file):
                # convert auth plugin paths (e.g.: gke-gcloud-auth-plugin)
                kubeconfig_file = kubernetes_utils.format_kubeconfig_exec_auth_with_cache(kubeconfig_file)  # pylint: disable=line-too-long

            # Upload kubeconfig to the default path to avoid having to set
            # KUBECONFIG in the environment.
            return {kubernetes.DEFAULT_KUBECONFIG_PATH: kubeconfig_file}
        else:
            return {}

    def instance_type_exists(self, instance_type: str) -> bool:
        return kubernetes_utils.KubernetesInstanceType.is_valid_instance_type(
            instance_type)

    def validate_region_zone(self, region: Optional[str], zone: Optional[str]):
        if region == kubernetes.in_cluster_context_name():
            # If running incluster, we set region to IN_CLUSTER_REGION
            # since there is no context name available.
            return region, zone

        all_contexts = kubernetes_utils.get_all_kube_context_names()

        if region and region not in all_contexts:
            raise ValueError(
                f'Context {region} not found in kubeconfig. Kubernetes only '
                'supports context names as regions. Available '
                f'contexts: {all_contexts}')
        if zone is not None:
            raise ValueError('Kubernetes support does not support setting zone.'
                             ' Cluster used is determined by the kubeconfig.')
        return region, zone

    @staticmethod
    def get_identity_from_context(context):
        if 'namespace' in context['context']:
            namespace = context['context']['namespace']
        else:
            namespace = kubernetes_utils.DEFAULT_NAMESPACE
        user = context['context']['user']
        cluster = context['context']['cluster']
        identity_str = f'{cluster}_{user}_{namespace}'
        return identity_str

    @classmethod
    def get_user_identities(cls) -> Optional[List[List[str]]]:
        identities = []
        k8s = kubernetes.kubernetes
        try:
            all_contexts, current_context = (
                kubernetes.list_kube_config_contexts())
        except k8s.config.config_exception.ConfigException:
            return None
        # Add current context at the head of the list
        current_identity = [cls.get_identity_from_context(current_context)]
        identities.append(current_identity)
        for context in all_contexts:
            identity = [cls.get_identity_from_context(context)]
            identities.append(identity)
        return identities

    @classmethod
    def is_volume_name_valid(cls,
                             volume_name: str) -> Tuple[bool, Optional[str]]:
        """Validates that the volume name is valid for this cloud.

        Follows Kubernetes DNS-1123 subdomain rules:
        - must be <= 253 characters
        - must match: '[a-z0-9]([-a-z0-9]*[a-z0-9])?(.[a-z0-9]([-a-z0-9]*[a-z0-9])?)*' # pylint: disable=line-too-long
        """
        # Max length per DNS-1123 subdomain
        if len(volume_name) > cls._MAX_VOLUME_NAME_LEN_LIMIT:
            return (False, f'Volume name exceeds the maximum length of '
                    f'{cls._MAX_VOLUME_NAME_LEN_LIMIT} characters '
                    '(DNS-1123 subdomain).')

        # DNS-1123 label: [a-z0-9]([-a-z0-9]*[a-z0-9])?
        label = r'[a-z0-9]([-a-z0-9]*[a-z0-9])?'
        # DNS-1123 subdomain: label(\.-separated label)*
        subdomain_pattern = rf'^{label}(\.{label})*$'
        if re.fullmatch(subdomain_pattern, volume_name) is None:
            return (False, 'Volume name must be a valid DNS-1123 subdomain: '
                    'lowercase alphanumeric, "-", and "."; start/end with '
                    'alphanumeric.')
        return True, None

    @classmethod
    def is_label_valid(cls, label_key: str,
                       label_value: str) -> Tuple[bool, Optional[str]]:
        # Kubernetes labels can be of the format <domain>/<key>: <value>
        key_regex = re.compile(
            # Look-ahead to ensure proper domain formatting up to a slash
            r'^(?:(?=[a-z0-9]([-a-z0-9.]*[a-z0-9])?\/)'
            # Match domain: starts and ends with alphanum up to 253 chars
            # including a slash in the domain.
            r'[a-z0-9]([-a-z0-9.]{0,251}[a-z0-9])?\/)?'
            # Match key: starts and ends with alphanum, upto to 63 chars.
            r'[a-z0-9]([-a-z0-9_.]{0,61}[a-z0-9])?$')
        value_regex = re.compile(
            r'^([a-zA-Z0-9]([-a-zA-Z0-9_.]{0,61}[a-zA-Z0-9])?)?$')
        key_valid = bool(key_regex.match(label_key))
        value_valid = bool(value_regex.match(label_value))
        error_msg = None
        condition_msg = ('Value must consist of alphanumeric characters or '
                         '\'-\', \'_\', \'.\', and must be no more than 63 '
                         'characters in length.')
        if not key_valid:
            error_msg = (f'Invalid label key {label_key} for Kubernetes. '
                         f'{condition_msg}')
        if not value_valid:
            error_msg = (f'Invalid label value {label_value} for Kubernetes. '
                         f'{condition_msg}')
        if not key_valid or not value_valid:
            return False, error_msg
        return True, None

    @classmethod
    def expand_infras(cls) -> List[str]:
        return [
            f'{cls.canonical_name()}/{c}'
            for c in cls.existing_allowed_contexts(silent=True)
        ]

    @classmethod
    def _detect_network_type(
        cls,
        context: str,
        network_tier: Optional['resources_utils.NetworkTier'] = None
    ) -> Tuple[KubernetesHighPerformanceNetworkType, str]:
        """Detect the type of Kubernetes network based on node labels.

        Args:
            context: The Kubernetes context to check.
            network_tier: The network tier requested. If None or not BEST,
                         returns NONE (no high-performance networking).

        Returns:
            A tuple of the detected network type and the instance type.
        """
        # If network_tier is None or not BEST, return NONE
        if (network_tier is None or
                network_tier != resources_utils.NetworkTier.BEST):
            return KubernetesHighPerformanceNetworkType.NONE, ''

        try:
            nodes = kubernetes_utils.get_kubernetes_nodes(context=context)
            for node in nodes:
                if node.metadata.labels:
                    # Check for Nebius clusters
                    for label_key, _ in node.metadata.labels.items():
                        if label_key.startswith('nebius.com/'):
                            return (KubernetesHighPerformanceNetworkType.NEBIUS,
                                    '')
                        if label_key.startswith('ib.coreweave.cloud/'):
                            return (
                                KubernetesHighPerformanceNetworkType.COREWEAVE,
                                '')

                    # Check for GKE clusters with specific GPUDirect variants
                    machine_family = node.metadata.labels.get(
                        'cloud.google.com/machine-family', '')
                    instance_type = node.metadata.labels.get(
                        'node.kubernetes.io/instance-type', '')
                    gke_accelerator = node.metadata.labels.get(
                        'cloud.google.com/gke-accelerator', '')

                    # Check if this is a GKE cluster with A3/A4 machine family
                    if machine_family in ['a3', 'a4']:
                        # Check instance type to determine specific GPUDirect
                        # variant
                        if 'a3-highgpu-8g' in instance_type:
                            return (
                                KubernetesHighPerformanceNetworkType.GCP_TCPX,
                                'a3-highgpu-8g')
                        elif 'a3-edgegpu-8g' in instance_type:
                            return (
                                KubernetesHighPerformanceNetworkType.GCP_TCPX,
                                'a3-edgegpu-8g')
                        elif 'a3-megagpu-8g' in instance_type:
                            return (
                                KubernetesHighPerformanceNetworkType.GCP_TCPXO,
                                'a3-megagpu-8g')
                        elif 'a4-highgpu-8g' in instance_type:
                            return (KubernetesHighPerformanceNetworkType.
                                    GCP_GPUDIRECT_RDMA, 'a4-highgpu-8g')
                        elif 'a3-ultragpu-8g' in instance_type:
                            return (KubernetesHighPerformanceNetworkType.
                                    GCP_GPUDIRECT_RDMA, 'a3-ultragpu-8g')
                        # Generic A3/A4 detection as fallback
                        elif machine_family == 'a4':
                            return (KubernetesHighPerformanceNetworkType.
                                    GCP_GPUDIRECT_RDMA, 'a4')

                    # Fallback: Check for GPU Direct TCPX capable instance
                    # types with high-perf GPUs
                    is_gpu_direct_tcpx_instance = (
                        instance_type
                        in gcp_constants.GPU_DIRECT_TCPX_INSTANCE_TYPES)
                    has_high_perf_gpu = ('nvidia-h100' in gke_accelerator or
                                         'nvidia-h200' in gke_accelerator or
                                         'nvidia-b200' in gke_accelerator)

                    if is_gpu_direct_tcpx_instance and has_high_perf_gpu:
                        # Default to TCPX if we can't determine the specific
                        # variant
                        return (KubernetesHighPerformanceNetworkType.GCP_TCPX,
                                instance_type)

        except exceptions.KubeAPIUnreachableError:
            # If we can't reach the cluster, assume no high perf networking
            pass

        # If we cannot determine the network type based on nodes
        # Check if the cluster has any node pools with autoscaling enabled
        # with machine types that support high perf networking for GKE.
        autoscaler_type = skypilot_config.get_effective_region_config(
            cloud='kubernetes',
            region=context,
            keys=('autoscaler',),
            default_value=None)
        if (autoscaler_type !=
                kubernetes_enums.KubernetesAutoscalerType.GKE.value):
            return KubernetesHighPerformanceNetworkType.NONE, ''
        autoscaler = kubernetes_utils.get_autoscaler(
            kubernetes_enums.KubernetesAutoscalerType(autoscaler_type))
        logger.debug(f'{context} has autoscaler of type: {autoscaler_type}')
        machine_types = autoscaler.get_available_machine_types(context)
        # Check if any machine type supports high perf networking for GKE.
        if 'a3-highgpu-8g' in machine_types:
            return (KubernetesHighPerformanceNetworkType.GCP_TCPX,
                    'a3-highgpu-8g')
        elif 'a3-edgegpu-8g' in machine_types:
            return (KubernetesHighPerformanceNetworkType.GCP_TCPX,
                    'a3-edgegpu-8g')
        elif 'a3-megagpu-8g' in machine_types:
            return (KubernetesHighPerformanceNetworkType.GCP_TCPXO,
                    'a3-megagpu-8g')
        elif 'a4-highgpu-8g' in machine_types:
            return (KubernetesHighPerformanceNetworkType.GCP_GPUDIRECT_RDMA,
                    'a4-highgpu-8g')
        elif 'a3-ultragpu-8g' in machine_types:
            return (KubernetesHighPerformanceNetworkType.GCP_GPUDIRECT_RDMA,
                    'a3-ultragpu-8g')

        return KubernetesHighPerformanceNetworkType.NONE, ''



"""Utilities for docker image generation."""
import os
import shutil
import subprocess
import tempfile
import textwrap
from typing import Dict, Optional, Tuple

import colorama

from sky import sky_logging
from sky import task as task_mod
from sky.adaptors import docker

logger = sky_logging.init_logger(__name__)

# Add docker-cli from official docker image to support docker-in-docker.
# We copy instead of installing docker-cli to keep the image builds fast.
DOCKERFILE_TEMPLATE = r"""
FROM {base_image}
SHELL ["/bin/bash", "-c"]
COPY --from=docker:dind /usr/local/bin/docker /usr/local/bin/
RUN apt-get update && apt-get -y install sudo
""".strip()

DOCKERFILE_SETUPCMD = """RUN {setup_command}"""
DOCKERFILE_COPYCMD = """COPY {copy_command}"""
DOCKERFILE_RUNCMD = """CMD {run_command}"""

CONDA_SETUP_PREFIX = '. $(conda info --base)/etc/profile.d/conda.sh 2> ' \
                     '/dev/null || true '

SKY_DOCKER_SETUP_SCRIPT = 'sky_setup.sh'
SKY_DOCKER_RUN_SCRIPT = 'sky_run.sh'
SKY_DOCKER_WORKDIR = 'sky_workdir'


def create_dockerfile(
    base_image: str,
    setup_command: Optional[str],
    copy_path: str,
    build_dir: str,
    run_command: Optional[str] = None,
) -> Tuple[str, Dict[str, str]]:
    """Writes a valid dockerfile to the specified path.

    performs three operations:
    1. load base_image
    2. run some setup commands
    3. copy a directory to the image.
    the run/entrypoint can be optionally specified in the dockerfile or at
    execution time.

    Args:
        base_image: base image to inherit from
        setup_command: commands to run for setup. eg. "pip install numpy && apt
            install htop"
        copy_path: local path to copy into the image. these are placed in the
            root of the container.
        build_dir: Where to write the dockerfile and setup scripts
        run_command: cmd argument to the dockerfile. optional - can also be
            specified at runtime.

    Returns:
        Tuple[dockerfile_contents(str), img_metadata(Dict[str, str])]
        dockerfile_contents: contents of the dockerfile
        img_metadata: metadata related to the image, propagated to the backend
    """
    img_metadata = {}
    dockerfile_contents = DOCKERFILE_TEMPLATE.format(base_image=base_image)

    # Copy workdir to image
    workdir_name = ''
    if copy_path:
        workdir_name = os.path.basename(os.path.dirname(copy_path))
        # NOTE: This relies on copy_path being copied to build context.
        copy_docker_cmd = f'{workdir_name} /{workdir_name}/'
        dockerfile_contents += '\n' + DOCKERFILE_COPYCMD.format(
            copy_command=copy_docker_cmd)

    def add_script_to_dockerfile(dockerfile_contents: str,
                                 multiline_cmds: Optional[str],
                                 out_filename: str):
        # Converts multiline commands to a script and adds the script to the
        # dockerfile. You still need to add the docker command to run the
        # script (either as CMD or RUN).
        script_path = os.path.join(build_dir, out_filename)
        bash_codegen(workdir_name, multiline_cmds, script_path)

        # Add CMD to run setup
        copy_cmd = f'{out_filename} /sky/{out_filename}'

        # Set permissions and add to dockerfile
        dockerfile_contents += '\n' + DOCKERFILE_COPYCMD.format(
            copy_command=copy_cmd)
        dockerfile_contents += '\n' + DOCKERFILE_SETUPCMD.format(
            setup_command=f'chmod +x ./sky/{out_filename}')
        return dockerfile_contents

    # ===== SETUP ======
    dockerfile_contents = add_script_to_dockerfile(dockerfile_contents,
                                                   setup_command,
                                                   SKY_DOCKER_SETUP_SCRIPT)
    cmd = f'./sky/{SKY_DOCKER_SETUP_SCRIPT}'
    dockerfile_contents += '\n' + DOCKERFILE_SETUPCMD.format(setup_command=cmd)

    # ===== RUN ======
    dockerfile_contents = add_script_to_dockerfile(dockerfile_contents,
                                                   run_command,
                                                   SKY_DOCKER_RUN_SCRIPT)
    cmd = f'./sky/{SKY_DOCKER_RUN_SCRIPT}'
    dockerfile_contents += '\n' + DOCKERFILE_RUNCMD.format(run_command=cmd)

    # Write Dockerfile
    with open(os.path.join(build_dir, 'Dockerfile'), 'w',
              encoding='utf-8') as f:
        f.write(dockerfile_contents)

    img_metadata['workdir_name'] = workdir_name
    return dockerfile_contents, img_metadata


def _execute_build(tag, context_path):
    """Executes a dockerfile build with the given context.

    The context path must contain the dockerfile and all dependencies.
    """
    assert tag is not None, 'Image tag cannot be None - have you specified a ' \
                            'task name? '
    docker_client = docker.from_env()
    try:
        unused_image, unused_build_logs = docker_client.images.build(
            path=context_path, tag=tag, rm=True, quiet=False)
    except docker.build_error() as e:
        style = colorama.Style
        fore = colorama.Fore
        logger.error(f'{fore.RED}Image build for {tag} failed - are your setup '
                     f'commands correct? Logs below{style.RESET_ALL}')
        logger.error(
            f'{style.BRIGHT}Image context is available at {context_path}'
            f'{style.RESET_ALL}')
        for line in e.build_log:
            if 'stream' in line:
                logger.error(line['stream'].strip())
        raise


def build_dockerimage(task: task_mod.Task,
                      tag: str) -> Tuple[str, Dict[str, str]]:
    """Builds a docker image for the given task.

    This method is responsible for:
    1. Create a temp directory to set the build context.
    2. Copy dockerfile to this directory and copy contents
    3. Run the dockerbuild
    """
    # Get tempdir
    temp_dir = tempfile.mkdtemp(prefix='sky_local_')

    # Create dockerfile
    if callable(task.run):
        raise ValueError(
            'Cannot build docker image for a task.run with function.')
    _, img_metadata = create_dockerfile(base_image=task.docker_image,
                                        setup_command=task.setup,
                                        copy_path=f'{SKY_DOCKER_WORKDIR}/',
                                        run_command=task.run,
                                        build_dir=temp_dir)

    dst = os.path.join(temp_dir, SKY_DOCKER_WORKDIR)
    if task.workdir is not None and isinstance(task.workdir, str):
        # Copy workdir contents to tempdir
        shutil.copytree(os.path.expanduser(task.workdir), dst)
    else:
        # Create an empty dir
        os.makedirs(dst)

    logger.info(f'Using tempdir {temp_dir} for docker build.')

    # Run docker image build
    _execute_build(tag, context_path=temp_dir)

    # Clean up temp dir
    subprocess.run(['rm', '-rf', temp_dir], check=False)

    return tag, img_metadata


def build_dockerimage_from_task(
        task: task_mod.Task) -> Tuple[str, Dict[str, str]]:
    """ Builds a docker image from a Task"""
    assert task.name is not None, task
    tag, img_metadata = build_dockerimage(task, tag=task.name)
    return tag, img_metadata


def push_dockerimage(local_tag, remote_name):
    raise NotImplementedError('Pushing images is not yet implemented.')


def make_bash_from_multiline(codegen: str) -> str:
    """Makes a bash script from a multi-line string of commands.

    Automatically includes conda setup prefixes.
    Args:
        codegen: str: multiline commands to be converted to a shell script

    Returns:
        script: str: str of shell script that can be written to a file
    """
    script = [
        textwrap.dedent(f"""\
        #!/bin/bash
        set -e
        {CONDA_SETUP_PREFIX}"""),
        codegen,
    ]
    script = '\n'.join(script)
    return script


def bash_codegen(workdir_name: str,
                 multiline_cmds: Optional[str],
                 out_path: Optional[str] = None):
    # Generate commands (if they exist) script and write to file
    if not multiline_cmds:
        multiline_cmds = ''
    multiline_cmds = f'cd /{workdir_name}\n{multiline_cmds}'
    script_contents = make_bash_from_multiline(multiline_cmds)
    if out_path:
        with open(out_path, 'w', encoding='utf-8') as fp:
            fp.write(script_contents)
    return script_contents



"""SkyServe core APIs."""
import typing
from typing import Any, Dict, List, Optional, Tuple, Union

from sky import backends
from sky import exceptions
from sky import sky_logging
from sky.backends import backend_utils
from sky.serve import serve_utils
from sky.serve.server import impl
from sky.usage import usage_lib
from sky.utils import controller_utils
from sky.utils import subprocess_utils

if typing.TYPE_CHECKING:
    import sky

logger = sky_logging.init_logger(__name__)


@usage_lib.entrypoint
def up(
    task: 'sky.Task',
    service_name: Optional[str] = None,
) -> Tuple[str, str]:
    """Spins up a service.

    Please refer to the sky.cli.serve_up for the document.

    Args:
        task: sky.Task to serve up.
        service_name: Name of the service.

    Returns:
        service_name: str; The name of the service.  Same if passed in as an
            argument.
        endpoint: str; The service endpoint.
    """
    return impl.up(task, service_name, pool=False)


@usage_lib.entrypoint
def update(
        task: 'sky.Task',
        service_name: str,
        mode: serve_utils.UpdateMode = serve_utils.DEFAULT_UPDATE_MODE) -> None:
    """Updates an existing service.

    Please refer to the sky.cli.serve_update for the document.

    Args:
        task: sky.Task to update.
        service_name: Name of the service.
        mode: Update mode.
    """
    return impl.update(task, service_name, mode, pool=False)


@usage_lib.entrypoint
# pylint: disable=redefined-builtin
def down(
    service_names: Optional[Union[str, List[str]]] = None,
    all: bool = False,
    purge: bool = False,
) -> None:
    """Tears down a service.

    Please refer to the sky.cli.serve_down for the docs.

    Args:
        service_names: Name of the service(s).
        all: Whether to terminate all services.
        purge: Whether to terminate services in a failed status. These services
          may potentially lead to resource leaks.

    Raises:
        sky.exceptions.ClusterNotUpError: if the sky serve controller is not up.
        ValueError: if the arguments are invalid.
        RuntimeError: if failed to terminate the service.
    """
    return impl.down(service_names, all, purge, pool=False)


@usage_lib.entrypoint
def terminate_replica(service_name: str, replica_id: int, purge: bool) -> None:
    """Tears down a specific replica for the given service.

    Args:
        service_name: Name of the service.
        replica_id: ID of replica to terminate.
        purge: Whether to terminate replicas in a failed status. These replicas
          may lead to resource leaks, so we require the user to explicitly
          specify this flag to make sure they are aware of this potential
          resource leak.

    Raises:
        sky.exceptions.ClusterNotUpError: if the sky sere controller is not up.
        RuntimeError: if failed to terminate the replica.
    """
    handle = backend_utils.is_controller_accessible(
        controller=controller_utils.Controllers.SKY_SERVE_CONTROLLER,
        stopped_message=
        'No service is running now. Please spin up a service first.',
        non_existent_message='No service is running now. '
        'Please spin up a service first.',
    )

    backend = backend_utils.get_backend_from_handle(handle)
    assert isinstance(backend, backends.CloudVmRayBackend)

    code = serve_utils.ServeCodeGen.terminate_replica(service_name, replica_id,
                                                      purge)
    returncode, stdout, stderr = backend.run_on_head(handle,
                                                     code,
                                                     require_outputs=True,
                                                     stream_logs=False,
                                                     separate_stderr=True)

    try:
        subprocess_utils.handle_returncode(returncode,
                                           code,
                                           'Failed to terminate the replica',
                                           stderr,
                                           stream_logs=True)
    except exceptions.CommandError as e:
        raise RuntimeError(e.error_msg) from e

    sky_logging.print(stdout)


@usage_lib.entrypoint
def status(
    service_names: Optional[Union[str,
                                  List[str]]] = None) -> List[Dict[str, Any]]:
    """Gets service statuses.

    If service_names is given, return those services. Otherwise, return all
    services.

    Each returned value has the following fields:

    .. code-block:: python

        {
            'name': (str) service name,
            'active_versions': (List[int]) a list of versions that are active,
            'controller_job_id': (int) the job id of the controller,
            'uptime': (int) uptime in seconds,
            'status': (sky.ServiceStatus) service status,
            'controller_port': (Optional[int]) controller port,
            'load_balancer_port': (Optional[int]) load balancer port,
            'endpoint': (Optional[str]) load balancer endpoint,
            'policy': (Optional[str]) autoscaling policy description,
            'requested_resources_str': (str) str representation of
              requested resources,
            'load_balancing_policy': (str) load balancing policy name,
            'tls_encrypted': (bool) whether the service is TLS encrypted,
            'replica_info': (List[Dict[str, Any]]) replica information,
        }

    Each entry in replica_info has the following fields:

    .. code-block:: python

        {
            'replica_id': (int) replica id,
            'name': (str) replica name,
            'status': (sky.serve.ReplicaStatus) replica status,
            'version': (int) replica version,
            'launched_at': (int) timestamp of launched,
            'handle': (ResourceHandle) handle of the replica cluster,
            'endpoint': (str) endpoint of the replica,
        }

    For possible service statuses and replica statuses, please refer to
    sky.cli.serve_status.

    Args:
        service_names: a single or a list of service names to query. If None,
            query all services.

    Returns:
        A list of dicts, with each dict containing the information of a service.
        If a service is not found, it will be omitted from the returned list.

    Raises:
        RuntimeError: if failed to get the service status.
        exceptions.ClusterNotUpError: if the sky serve controller is not up.
    """
    return impl.status(service_names, pool=False)


ServiceComponentOrStr = Union[str, serve_utils.ServiceComponent]


@usage_lib.entrypoint
def tail_logs(
    service_name: str,
    *,
    target: ServiceComponentOrStr,
    replica_id: Optional[int] = None,
    follow: bool = True,
    tail: Optional[int] = None,
) -> None:
    """Tails logs for a service.

    Usage:
        sky.serve.tail_logs(
            service_name,
            target=<component>,
            follow=False, # Optionally, default to True
            # replica_id=3, # Must be specified when target is REPLICA.
        )

    `target` is a enum of sky.serve.ServiceComponent, which can be one of:
        - CONTROLLER
        - LOAD_BALANCER
        - REPLICA
    Pass target as a lower-case string is also supported, e.g.
    target='controller'.
    To use REPLICA, you must specify `replica_id`.

    To tail controller logs:
        # follow default to True
        sky.serve.tail_logs(
            service_name, target=sky.serve.ServiceComponent.CONTROLLER)

    To print replica 3 logs:
        # Pass target as a lower-case string is also supported.
        sky.serve.tail_logs(
            service_name, target='replica',
            follow=False, replica_id=3)

    Raises:
        sky.exceptions.ClusterNotUpError: the sky serve controller is not up.
        ValueError: arguments not valid, or failed to tail the logs.
    """
    return impl.tail_logs(service_name,
                          target=target,
                          replica_id=replica_id,
                          follow=follow,
                          tail=tail,
                          pool=False)


@usage_lib.entrypoint
def sync_down_logs(
    service_name: str,
    *,
    local_dir: str,
    targets: Union[ServiceComponentOrStr, List[ServiceComponentOrStr],
                   None] = None,
    replica_ids: Optional[List[int]] = None,
    tail: Optional[int] = None,
) -> str:
    """Sync down logs from the controller for the given service.

    This function is called by the server endpoint. It gathers logs from the
    controller, load balancer, and/or replicas and places them in a directory
    under the user's log space on the API server filesystem.

    Args:
        service_name: The name of the service to download logs from.
        local_dir: The local directory to save the logs to.
        targets: Which component(s) to download logs for. If None or empty,
            means download all logs (controller, load-balancer, all replicas).
            Can be a string (e.g. "controller"), or a `ServiceComponent` object,
            or a list of them for multiple components. Currently accepted
            values:
                - "controller"/ServiceComponent.CONTROLLER
                - "load_balancer"/ServiceComponent.LOAD_BALANCER
                - "replica"/ServiceComponent.REPLICA
        replica_ids: The list of replica IDs to download logs from, specified
            when target includes `ServiceComponent.REPLICA`. If target includes
            `ServiceComponent.REPLICA` but this is None/empty, logs for all
            replicas will be downloaded.

    Returns:
        A dict mapping component names to local paths where the logs were synced
        down to.

    Raises:
        RuntimeError: If fails to gather logs or fails to rsync from the
          controller.
        sky.exceptions.ClusterNotUpError: If the controller is not up.
        ValueError: Arguments not valid.
    """
    return impl.sync_down_logs(service_name,
                               local_dir=local_dir,
                               targets=targets,
                               replica_ids=replica_ids,
                               tail=tail,
                               pool=False)



"""Encoders for the REST API return values."""
# TODO(SKY-1211): we should evaluate that if we can move our return values to
# pydantic models, so we can take advantage of model_dump_json of pydantic,
# instead of implementing our own handlers.
import base64
import dataclasses
import pickle
import typing
from typing import Any, Dict, List, Optional, Tuple, Union

from sky.schemas.api import responses
from sky.server import constants as server_constants
from sky.utils import serialize_utils

if typing.TYPE_CHECKING:
    from sky import backends
    from sky import clouds
    from sky import models
    from sky.provision.kubernetes import utils as kubernetes_utils

handlers: Dict[str, Any] = {}


def pickle_and_encode(obj: Any) -> str:
    try:
        # Apply backwards compatibility processing at the lowest level
        # to catch any handles that might have bypassed the encoders
        obj = serialize_utils.prepare_handle_for_backwards_compatibility(obj)
        return base64.b64encode(pickle.dumps(obj)).decode('utf-8')
    except TypeError as e:
        raise ValueError(f'Failed to pickle object: {obj}') from e


def register_encoder(*names: str):
    """Decorator to register an encoder."""

    def decorator(func):
        for name in names:
            if name != server_constants.DEFAULT_HANDLER_NAME:
                name = server_constants.REQUEST_NAME_PREFIX + name
            handlers[name] = func
        return func

    return decorator


def get_encoder(name: str):
    """Get the encoder for a request name."""
    return handlers.get(name, handlers[server_constants.DEFAULT_HANDLER_NAME])


@register_encoder(server_constants.DEFAULT_HANDLER_NAME)
def default_encoder(return_value: Any) -> Any:
    """The default encoder."""
    return return_value


@register_encoder('status')
def encode_status(
        clusters: List[responses.StatusResponse]) -> List[Dict[str, Any]]:
    response = []
    for cluster in clusters:
        response_cluster = cluster.model_dump()
        response_cluster['status'] = cluster['status'].value
        handle = serialize_utils.prepare_handle_for_backwards_compatibility(
            cluster['handle'])
        response_cluster['handle'] = pickle_and_encode(handle)
        response_cluster['storage_mounts_metadata'] = pickle_and_encode(
            response_cluster['storage_mounts_metadata'])
        response.append(response_cluster)
    return response


@register_encoder('launch', 'exec', 'jobs.launch')
def encode_launch(
    job_id_handle: Tuple[Optional[int], Optional['backends.ResourceHandle']]
) -> Dict[str, Any]:
    job_id, handle = job_id_handle
    handle = serialize_utils.prepare_handle_for_backwards_compatibility(handle)
    return {
        'job_id': job_id,
        'handle': pickle_and_encode(handle),
    }


@register_encoder('start')
def encode_start(resource_handle: 'backends.CloudVmRayResourceHandle') -> str:
    resource_handle = (
        serialize_utils.prepare_handle_for_backwards_compatibility(
            resource_handle))
    return pickle_and_encode(resource_handle)


@register_encoder('queue')
def encode_queue(jobs: List[dict],) -> List[Dict[str, Any]]:
    for job in jobs:
        job['status'] = job['status'].value
    return jobs


@register_encoder('status_kubernetes')
def encode_status_kubernetes(
    return_value: Tuple[
        List['kubernetes_utils.KubernetesSkyPilotClusterInfoPayload'],
        List['kubernetes_utils.KubernetesSkyPilotClusterInfoPayload'],
        List[Dict[str, Any]], Optional[str]]
) -> Tuple[List[Dict[str, Any]], List[Dict[str, Any]], List[Dict[str, Any]],
           Optional[str]]:
    all_clusters, unmanaged_clusters, all_jobs, context = return_value
    encoded_all_clusters = []
    encoded_unmanaged_clusters = []
    for cluster in all_clusters:
        encoded_cluster = dataclasses.asdict(cluster)
        encoded_cluster['status'] = encoded_cluster['status'].value
        encoded_all_clusters.append(encoded_cluster)
    for cluster in unmanaged_clusters:
        encoded_cluster = dataclasses.asdict(cluster)
        encoded_cluster['status'] = encoded_cluster['status'].value
        encoded_unmanaged_clusters.append(encoded_cluster)
    return encoded_all_clusters, encoded_unmanaged_clusters, all_jobs, context


@register_encoder('jobs.queue')
def encode_jobs_queue(jobs: List[dict],) -> List[Dict[str, Any]]:
    for job in jobs:
        job['status'] = job['status'].value
    return jobs


@register_encoder('jobs.queue_v2')
def encode_jobs_queue_v2(
        jobs_or_tuple) -> Union[List[Dict[str, Any]], Dict[str, Any]]:
    # Support returning either a plain jobs list or a (jobs, total) tuple
    status_counts: Dict[str, int] = {}
    if isinstance(jobs_or_tuple, tuple):
        if len(jobs_or_tuple) == 2:
            jobs, total = jobs_or_tuple
            total_no_filter = total
        elif len(jobs_or_tuple) == 4:
            jobs, total, status_counts, total_no_filter = jobs_or_tuple
        else:
            raise ValueError(f'Invalid jobs tuple: {jobs_or_tuple}')
    else:
        jobs = jobs_or_tuple
        total = None
    for job in jobs:
        job['status'] = job['status'].value
    if total is None:
        return jobs
    return {
        'jobs': jobs,
        'total': total,
        'total_no_filter': total_no_filter,
        'status_counts': status_counts
    }


def _encode_serve_status(
        service_statuses: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    for service_status in service_statuses:
        service_status['status'] = service_status['status'].value
        for replica_info in service_status.get('replica_info', []):
            replica_info['status'] = replica_info['status'].value
            handle = serialize_utils.prepare_handle_for_backwards_compatibility(
                replica_info['handle'])
            replica_info['handle'] = pickle_and_encode(handle)
    return service_statuses


@register_encoder('serve.status')
def encode_serve_status(
        service_statuses: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    return _encode_serve_status(service_statuses)


@register_encoder('jobs.pool_status')
def encode_jobs_pool_status(
        pool_statuses: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    return _encode_serve_status(pool_statuses)


@register_encoder('cost_report')
def encode_cost_report(
        cost_report: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    for cluster_report in cost_report:
        if cluster_report['status'] is not None:
            cluster_report['status'] = cluster_report['status'].value
        cluster_report['resources'] = pickle_and_encode(
            cluster_report['resources'])
    return cost_report


@register_encoder('enabled_clouds')
def encode_enabled_clouds(clouds: List['clouds.Cloud']) -> List[str]:
    enabled_clodus_list = [str(cloud) for cloud in clouds]
    return enabled_clodus_list


@register_encoder('storage_ls')
def encode_storage_ls(
        return_value: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    for storage_info in return_value:
        storage_info['status'] = storage_info['status'].value
        storage_info['store'] = [store.value for store in storage_info['store']]
    return return_value


@register_encoder('job_status')
def encode_job_status(return_value: Dict[int, Any]) -> Dict[int, str]:
    for job_id in return_value.keys():
        if return_value[job_id] is not None:
            return_value[job_id] = return_value[job_id].value
    return return_value


@register_encoder('kubernetes_node_info')
def encode_kubernetes_node_info(
        return_value: 'models.KubernetesNodesInfo') -> Dict[str, Any]:
    return return_value.to_dict()


@register_encoder('endpoints')
def encode_endpoints(return_value: Dict[int, str]) -> Dict[str, str]:
    return {str(k): v for k, v in return_value.items()}
